from dataclasses import dataclass
from typing import Literal, TypeAlias

import numpy as np
import numpy.typing as npt

from .. import abc
from .._error_control_block.PolarCode import PolarCode
from .._util.decorators import blockwise, vectorize
from .._util.special_functions import boxplus

Belief: TypeAlias = npt.NDArray[np.floating]
Decision: TypeAlias = npt.NDArray[np.integer]
Node: TypeAlias = tuple[int, int]  # depth, index


def g(r: Belief, s: Belief, b: Decision) -> Belief:
    return s + (-1) ** b * r


@dataclass
class SCDecoder(abc.BlockDecoder[PolarCode]):
    r"""
    Successive cancellation decoder for [Polar codes](/ref/PolarCode).

    Parameters:
        code: The Polar code to be used for decoding.
        output_type: The type of the output. Either `'hard'` or `'soft'`. Default is `'soft'`.

    Notes:
        - Input type: `soft` (L-values).
        - Output type: `hard` (bits) or `soft` (L-values).
    """

    code: PolarCode
    output_type: Literal["hard", "soft"] = "soft"

    def __post_init__(self) -> None:
        self._f = boxplus

    def decode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer | np.floating]:
        r"""
        Examples:
            >>> code = komm.PolarCode(3, [0, 1, 2, 4])

            >>> decoder = komm.SCDecoder(code)
            >>> decoder.decode([1, -4, -3, 2, -2, 3, 4, -1])
            array([ -6.84595089,  -5.96379094,  -9.30685282, -20.        ])

            >>> decoder = komm.SCDecoder(code, output_type="hard")
            >>> decoder.decode([1, -4, -3, 2, -2, 3, 4, -1])
            array([1, 1, 1, 1])
        """

        @blockwise(self.code.length)
        @vectorize
        def decode(li: npt.NDArray[np.floating]) -> npt.NDArray[np.floating]:
            # See [Successive Cancellation(SC) Decoder for a General (N,K) Polar Code]
            # by Prof. Andrew Thangaraj (NPTEL-NOC IITM)
            active = (0, 0)
            beliefs: dict[Node, Belief] = {(0, 0): li}
            decisions: dict[Node, Decision] = {}
            while active[0] >= 0:
                depth, index = active
                parent = (depth - 1, index // 2)
                if depth == self.code.mu:  # Leaf node
                    if index in self.code.frozen:
                        decisions[active] = np.array([0])
                    else:
                        decisions[active] = (beliefs[active] < 0).astype(int)
                    active = parent
                    continue
                # Interior node
                child_l = (depth + 1, index * 2)
                child_r = (depth + 1, index * 2 + 1)
                msg = beliefs[active]
                M = msg.size
                if child_l not in decisions:  # Step L:
                    beliefs[child_l] = self._f(msg[: M // 2], msg[M // 2 :])
                    active = child_l
                elif child_r not in decisions:  # Step R:
                    side_msg = decisions[child_l]
                    beliefs[child_r] = g(msg[: M // 2], msg[M // 2 :], side_msg)
                    active = child_r
                else:  # Step U:
                    a = decisions[child_l]
                    b = decisions[child_r]
                    decisions[active] = np.concatenate([a ^ b, b])
                    active = parent
            lo = np.array([beliefs[self.code.mu, i][0] for i in self.code.active])
            return lo

        output = decode(input)
        if self.output_type == "hard":
            return (output < 0).astype(int)
        else:
            return output



from math import ceil, log
from typing import Literal

import numpy as np
import numpy.typing as npt
from tqdm import tqdm

from .._util.validators import validate_integer_range
from .util import integer_to_symbols, symbols_to_integer

Token = tuple[Literal[0], int] | tuple[Literal[1], int, int]


class LempelZivSSCode:
    r"""
    Lempel–Ziv–Storer–Szymanski (LZSS) code. It is a lossless data compression algorithm which is a variation of the [Lempel–Ziv 77](/ref/LempelZiv77Code) algorithm. Let $\mathcal{X}$ be the source alphabet, $\mathcal{Y}$ be the target alphabet, $S \geq 1$ be the size of the *search buffer*, and $L \geq 1$ be the size of the *lookahead buffer*. The token format follows <cite>CT06, Sec. 13.4.1</cite>, where a token is either a *literal* $(0, x)$, where $x \in \mathcal{X}$ is a source symbol, or a *reference* $(1, p, \ell)$, where $p \in [1 : S]$ is the *pointer* (location of the beginning of the match, measured backward from the end of the search window), and $\ell \in [1 : L]$ is the *length* of the match. References are only encoded if they provide compression benefit (length exceeds the break-even point). For more details, see <cite>CT06, Sec. 13.4.1</cite>.

    Note:
        Here, for simplicity, we assume that the source alphabet is $\mathcal{X} = [0 : |\mathcal{X}|)$ and the target alphabet is $\mathcal{Y} = [0 : |\mathcal{Y}|)$, where $|\mathcal{X}| \geq 2$ and $|\mathcal{Y}| \geq 2$ are called the *source cardinality* and *target cardinality*, respectively.

    Parameters:
        search_size: The search buffer size $S$. Must satisfy $S \geq 1$.
        lookahead_size: The lookahead buffer size $L$. Must satisfy $L \geq 1$.
        source_cardinality: The source cardinality $|\mathcal{X}|$. Must satisfy $|\mathcal{X}| \geq 2$.
        target_cardinality: The target cardinality $|\mathcal{Y}|$. Must satisfy $|\mathcal{Y}| \geq 2$. The default value is $2$ (binary).
        search_buffer: The initial state of the search buffer. Must be a 1D-array of length $S$ with elements in $\mathcal{X}$. The default value corresponds to $(0, \ldots, 0) \in \mathcal{X}^S$.

    Examples:
        >>> lzss = komm.LempelZivSSCode(
        ...     search_size=2**13,
        ...     lookahead_size=16,
        ...     source_cardinality=256,
        ...     target_cardinality=2,
        ... )
    """

    def __init__(
        self,
        *,
        search_size: int,
        lookahead_size: int,
        source_cardinality: int,
        target_cardinality: int = 2,
        search_buffer: npt.ArrayLike | None = None,
    ):
        if not search_size >= 1:
            raise ValueError("'search_size' must be at least 1")
        if not lookahead_size >= 1:
            raise ValueError("'lookahead_size' must be at least 1")
        if not source_cardinality >= 2:
            raise ValueError("'source_cardinality' must be at least 2")
        if not target_cardinality >= 2:
            raise ValueError("'target_cardinality' must be at least 2")

        self.search_size = search_size
        self.lookahead_size = lookahead_size
        self.source_cardinality = source_cardinality
        self.target_cardinality = target_cardinality

        if search_buffer is None:
            self.search_buffer = [0] * self.search_size
        else:
            self.search_buffer = np.asarray(search_buffer, dtype=int).tolist()
            if len(self.search_buffer) != self.search_size:
                raise ValueError(
                    "length of 'search_buffer' must be S ="
                    f" {self.search_size} (got {len(self.search_buffer)})"
                )

    def __repr__(self) -> str:
        args = ", ".join([
            f"search_size={self.search_size}",
            f"lookahead_size={self.lookahead_size}",
            f"source_cardinality={self.source_cardinality}",
            f"target_cardinality={self.target_cardinality}",
        ])
        return f"{self.__class__.__name__}({args})"

    @property
    def window_size(self) -> int:
        r"""
        The sliding window size $W$. It is given by $W = S + L$.
        """
        return self.search_size + self.lookahead_size

    def _get_widths(self) -> tuple[int, int, int]:
        calY = self.target_cardinality
        p_width = ceil(log(self.search_size, calY))
        l_width = ceil(log(self.lookahead_size, calY))
        x_width = ceil(log(self.source_cardinality, calY))
        return p_width, l_width, x_width

    @property
    def break_even(self) -> int:
        r"""
        The break-even point for encoding a reference. A match must be at least this long to be worth encoding as a reference instead of literals. It is given by
        $$
            \left\lceil \frac{1 + \lceil \log S \rceil + \lceil \log L \rceil}{1+ \lceil \log |\mathcal{X}| \rceil} \right\rceil,
        $$
        where all logs are to base $|\mathcal{Y}|$.
        """
        p_width, l_width, x_width = self._get_widths()
        # Cost as a reference: 1 (flag) + p_width + l_width
        # Cost as literals: length * (1 (flag) + x_width)
        # Break even when: 1 + p_width + l_width < length * (1 + x_width)
        return ceil((1 + p_width + l_width) / (1 + x_width))

    def source_to_tokens(self, source: npt.ArrayLike) -> list[Token]:
        r"""
        Encodes a given sequence of source symbols to the corresponding list of tokens.

        Examples:
            >>> lzss = komm.LempelZivSSCode(
            ...     search_size=8,
            ...     lookahead_size=4,
            ...     source_cardinality=4,
            ... )
            >>> lzss.source_to_tokens([3, 0, 0, 0, 3, 2, 0, 3])
            [(0, 3), (1, 4, 4), (0, 2), (1, 3, 2)]
        """
        source = validate_integer_range(source, high=self.source_cardinality)
        ss, ls = self.search_size, self.lookahead_size
        break_even = self.break_even
        buffer = bytes(self.search_buffer + source.tolist())
        tokens: list[Token] = []

        pbar = tqdm(total=len(buffer), initial=ss, desc="Compressing LZSS", delay=2.5)
        i = ss
        while i < len(buffer):
            l = min(ls, len(buffer) - i)
            while True:
                p = buffer[i - ss : i + l - 1].rfind(buffer[i : i + l])
                if p >= 0:
                    break
                l -= 1
            if l >= break_even:  # Reference
                tokens.append((1, ss - p, l))
                i += l
                pbar.update(l)
            else:  # Literal
                tokens.append((0, buffer[i]))
                i += 1
                pbar.update(1)

        pbar.close()
        return tokens

    def tokens_to_source(self, tokens: list[Token]) -> npt.NDArray[np.integer]:
        r"""
        Decodes a given list of tokens to the corresponding sequence of source symbols.

        Examples:
            >>> lzss = komm.LempelZivSSCode(
            ...     search_size=8,
            ...     lookahead_size=4,
            ...     source_cardinality=4,
            ... )
            >>> lzss.tokens_to_source([(0, 3), (1, 4, 4), (0, 2), (1, 3, 2)])
            array([3, 0, 0, 0, 3, 2, 0, 3])
        """
        ss = self.search_size
        buffer = self.search_buffer.copy()

        for token in tokens:
            match token:
                case (1, p, l):  # Reference
                    for _ in range(l):
                        buffer.append(buffer[-p])
                case (0, x):  # Literal
                    buffer.append(x)
                case _:  # pyright: ignore[reportUnnecessaryComparison]
                    raise ValueError(f"invalid token: {token}")

        source = np.array(buffer[ss:], dtype=int)
        return source

    def tokens_to_target(self, tokens: list[Token]) -> npt.NDArray[np.integer]:
        r"""
        Returns the target alphabet representation corresponding to a given list of tokens.

        Examples:
            >>> lzss = komm.LempelZivSSCode(
            ...     search_size=8,
            ...     lookahead_size=4,
            ...     source_cardinality=4,
            ... )
            >>> lzss.tokens_to_target([(0, 3), (1, 4, 4), (0, 2), (1, 3, 2)])
            array([0, 1, 1, 1, 0, 1, 1, 1, 1, 0, 1, 0, 1, 0, 1, 0, 0, 1])
        """
        calY = self.target_cardinality
        p_width, l_width, x_width = self._get_widths()
        target: list[int] = []

        for token in tokens:
            match token:
                case (1, p, l):  # Reference
                    target.append(1)
                    target.extend(integer_to_symbols(p - 1, base=calY, width=p_width))
                    target.extend(integer_to_symbols(l - 1, base=calY, width=l_width))
                case (0, x):  # Literal
                    target.append(0)
                    target.extend(integer_to_symbols(x, base=calY, width=x_width))
                case _:  # pyright: ignore[reportUnnecessaryComparison]
                    raise ValueError(f"invalid token: {token}")

        return np.array(target, dtype=int)

    def target_to_tokens(self, target: npt.ArrayLike) -> list[Token]:
        r"""
        Returns the list of tokens corresponding to a given target alphabet representation.

        Examples:
            >>> lzss = komm.LempelZivSSCode(
            ...     search_size=8,
            ...     lookahead_size=4,
            ...     source_cardinality=4,
            ... )
            >>> lzss.target_to_tokens([0, 1, 1, 1, 0, 1, 1, 1, 1, 0, 1, 0, 1, 0, 1, 0, 0, 1])
            [(0, 3), (1, 4, 4), (0, 2), (1, 3, 2)]
        """
        target = np.asarray(target, dtype=int)
        calY = self.target_cardinality
        p_width, l_width, x_width = self._get_widths()
        tokens: list[Token] = []
        i = 0

        while i < target.size:
            flag = target[i]
            i += 1
            if flag == 1:  # Reference
                if i + p_width + l_width <= target.size:
                    p = int(symbols_to_integer(target[i : i + p_width], base=calY)) + 1
                    i += p_width
                    l = int(symbols_to_integer(target[i : i + l_width], base=calY)) + 1
                    i += l_width
                    tokens.append((1, p, l))
            else:  # Literal
                if i + x_width <= target.size:
                    x = int(symbols_to_integer(target[i : i + x_width], base=calY))
                    i += x_width
                    tokens.append((0, x))

        return tokens

    def encode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Encodes a sequence of source symbols to a sequence of target symbols.

        Parameters:
            input: The sequence of source symbols to be encoded. Must be a 1D-array with elements in $\mathcal{X}$.

        Returns:
            output: The sequence of encoded target symbols. It is a 1D-array with elements in $\mathcal{Y}$.

        Examples:
            >>> lzss = komm.LempelZivSSCode(
            ...     search_size=8,
            ...     lookahead_size=4,
            ...     source_cardinality=4,
            ... )
            >>> lzss.encode([3, 0, 0, 0, 3, 2, 0, 3])
            array([0, 1, 1, 1, 0, 1, 1, 1, 1, 0, 1, 0, 1, 0, 1, 0, 0, 1])
        """
        tokens = self.source_to_tokens(input)
        output = self.tokens_to_target(tokens)
        return output

    def decode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Decodes a sequence of target symbols to a sequence of source symbols.

        Parameters:
            input: The sequence of target symbols to be decoded. Must be a 1D-array with elements in $\mathcal{Y}$. Also, the sequence must be a valid output of the `encode` method.

        Returns:
            output: The sequence of decoded source symbols. It is a 1D-array with elements in $\mathcal{X}$.

        Examples:
            >>> lzss = komm.LempelZivSSCode(
            ...     search_size=8,
            ...     lookahead_size=4,
            ...     source_cardinality=4,
            ... )
            >>> lzss.decode([0, 1, 1, 1, 0, 1, 1, 1, 1, 0, 1, 0, 1, 0, 1, 0, 0, 1])
            array([3, 0, 0, 0, 3, 2, 0, 3])
        """
        tokens = self.target_to_tokens(input)
        output = self.tokens_to_source(tokens)
        return output



from dataclasses import dataclass
from typing import Literal

import numpy as np
import numpy.typing as npt

from .. import abc
from .._error_control_block.ReedMullerCode import ReedMullerCode
from .._util.decorators import blockwise, vectorize


@dataclass
class ReedDecoder(abc.BlockDecoder[ReedMullerCode]):
    r"""
    Reed decoder for [Reed-Muller codes](/ref/ReedMullerCode). It's a majority-logic decoding algorithm. For more details, see [LC04, Sec 4.3 and 10.9.1] for hard-decision decoding, and [LC04, Sec 10.9.2] for soft-decision decoding.

    Parameters:
        code: The Reed-Muller code to be used for decoding.
        input_type: The type of the input. Either `'hard'` or `'soft'`. Default is `'hard'`.

    Notes:
        - Input type: `hard` (bits) or `soft` (either L-values or channel outputs).
        - Output type: `hard` (bits).
    """

    code: ReedMullerCode
    input_type: Literal["hard", "soft"] = "hard"

    def __post_init__(self) -> None:
        self._reed_partitions = self.code.reed_partitions()

    def decode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer | np.floating]:
        r"""
        Examples:
            >>> code = komm.ReedMullerCode(1, 3)

            >>> decoder = komm.ReedDecoder(code, input_type="hard")
            >>> decoder.decode([[0, 0, 0, 0, 0, 1, 0, 0], [1, 1, 0, 1, 1, 1, 1, 1]])
            array([[0, 0, 0, 0],
                   [0, 0, 0, 1]])

            >>> decoder = komm.ReedDecoder(code, input_type="soft")
            >>> decoder.decode([+1.3, +1.0, +0.9, +0.4, -0.8, +0.2, +0.3, +0.8])
            array([0, 0, 0, 0])
        """

        @blockwise(self.code.length)
        @vectorize
        def decode_hard(r: npt.NDArray[np.integer]):
            u_hat = np.empty(self.code.dimension, dtype=int)
            bx = r.copy()
            for i, partition in enumerate(self._reed_partitions):
                checksums = np.count_nonzero(bx[partition], axis=1) % 2
                u_hat[i] = np.count_nonzero(checksums) > len(checksums) // 2
                bx ^= u_hat[i] * self.code.generator_matrix[i]
            return u_hat

        @blockwise(self.code.length)
        @vectorize
        def decode_soft(r: npt.NDArray[np.floating]):
            u_hat = np.empty(self.code.dimension, dtype=int)
            bx = (r < 0).astype(int)
            for i, partition in enumerate(self._reed_partitions):
                checksums = np.count_nonzero(bx[partition], axis=1) % 2
                min_reliability = np.min(np.abs(r[partition]), axis=1)
                decision_var = (1 - 2 * checksums) @ min_reliability
                u_hat[i] = decision_var < 0
                bx ^= u_hat[i] * self.code.generator_matrix[i]
            return u_hat

        if self.input_type == "hard":
            return decode_hard(input)
        else:  # self.input_type == "soft"
            return decode_soft(input)



from functools import cache

import numpy as np
import numpy.typing as npt

from .. import abc


class FibonacciCode(abc.IntegerCode):
    r"""
    Fibonacci code. It is an integer code. For the definition of this code, see [Wikipedia: Fibonacci coding](https://en.wikipedia.org/wiki/Fibonacci_coding).
    """

    def encode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> code = komm.FibonacciCode()
            >>> code.encode([4, 1, 3])
            array([1, 0, 1, 1, 1, 1, 0, 0, 1, 1])
        """
        input = np.asarray(input)
        if input.size == 0:
            return np.array([], dtype=int)
        return np.concatenate([fibonacci_encode(i) for i in input])

    def decode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> code = komm.FibonacciCode()
            >>> code.decode([1, 0, 1, 1, 1, 1, 0, 0, 1, 1])
            array([4, 1, 3])
        """
        input = np.asarray(input)
        output: list[int] = []
        i = 0
        while i < input.size:
            j = i
            while not input[j] == input[j + 1] == 1:
                j += 1
            output.append(fibonacci_decode(list(input[i : j + 2])))
            i = j + 2
        return np.array(output)


@cache
def fibonacci(n: int) -> int:
    if n == 0 or n == 1:
        return n
    return fibonacci(n - 1) + fibonacci(n - 2)


def fibonacci_encode(integer: int) -> list[int]:
    positions: list[int] = []
    while integer > 0:
        i = 0
        while fibonacci(i + 1) <= integer:
            i += 1
        integer -= fibonacci(i)
        positions.append(i - 2)
    last = max(positions) + 1
    return [1 if i in positions else 0 for i in range(last)] + [1]


def fibonacci_decode(bits: list[int]) -> int:
    return sum(fibonacci(i + 2) for i, bit in enumerate(bits[:-1]) if bit)



from collections.abc import Sequence
from typing import Any, Protocol, TypeVar, runtime_checkable

import numpy as np
import numpy.typing as npt
from typing_extensions import Self

T_co = TypeVar("T_co", bound="RingElement", covariant=True)


@runtime_checkable
class RingElement(Protocol):
    @property
    def ambient(self) -> "Ring[Any]": ...
    def __add__(self: Self, other: Self) -> Self: ...
    def __sub__(self: Self, other: Self) -> Self: ...
    def __neg__(self) -> Self: ...
    def __mul__(self: Self, other: Self) -> Self: ...
    def __rmul__(self: Self, other: int) -> Self: ...


@runtime_checkable
class Ring(Protocol[T_co]):
    def __call__(self, value: Any) -> T_co: ...
    def __eq__(self, other: Any) -> bool: ...
    @property
    def zero(self) -> T_co: ...
    @property
    def one(self) -> T_co: ...


def power(x: T_co, n: int) -> T_co:
    r"""
    Computes $x^n$ using exponentiation by squaring.

    Parameters:
        x: The base (a ring element)
        n: The exponent (a non-negative integer)

    Returns:
        power: The result of `x` raised to the power of `n` in the ring

    Raises:
        ValueError: If `n` is negative

    References:
        `Exponentiation by squaring <https://en.wikipedia.org/wiki/Exponentiation_by_squaring>`_
    """
    if n < 0:
        raise ValueError("Negative exponents not supported")
    if n == 0:
        return x.ambient.one
    if n == 1:
        return x
    y = power(x, n // 2)
    y = y * y
    if n % 2 == 1:
        y = x * y
    return y


def binary_horner(coefficients: npt.ArrayLike, x: T_co) -> T_co:
    r"""
    Specialized Horner's method for binary polynomials. See :func:`horner` for a more general implementation.
    """
    coefficients = np.asarray(coefficients)
    zero: T_co = x.ambient.zero
    one: T_co = x.ambient.one
    result = zero
    for coefficient in reversed(coefficients):
        result *= x
        if coefficient:
            result += coefficient * one
    return result


def horner(coefficients: npt.ArrayLike, x: T_co) -> T_co:
    r"""
    Evaluates a polynomial at a point using Horner's method.

    Parameters:
        poly: Polynomial to evaluate
        x: Point at which to evaluate the polynomial (ring element)

    Returns:
        result: The result of evaluating the polynomial `poly` at point `x`

    References:
        `Horner's method <https://en.wikipedia.org/wiki/Horner's_method>`_
    """
    coefficients = np.asarray(coefficients)
    zero: T_co = x.ambient.zero
    one: T_co = x.ambient.one
    result = zero
    for coefficient in reversed(coefficients):
        result = result * x + coefficient * one
    return result


def determinant(matrix: Sequence[Sequence[T_co]]) -> T_co:
    r"""
    Computes the determinant of a matrix using cofactor expansion.

    Parameters:
        matrix: The matrix with coefficients over a ring

    Returns:
        det: The determinant of the matrix (a ring element)

    References:
        `Laplace expansion <https://en.wikipedia.org/wiki/Laplace_expansion>`_
    """
    n = len(matrix)
    if any(len(row) != n for row in matrix):
        raise ValueError("'matrix' should be square")
    if n == 1:
        return matrix[0][0]
    zero: T_co = matrix[0][0].ambient.zero
    det = zero
    for j in range(n):
        minor = [list(row[:j]) + list(row[j + 1 :]) for row in matrix[1:]]
        sign = -1 if j % 2 else 1
        cofactor = matrix[0][j]
        det += (sign * cofactor) * determinant(minor)
    return det



from typing import TypeVar

import numpy as np
import numpy.typing as npt

from .. import abc
from ..types import Array1D, Array2D

T = TypeVar("T", np.floating, np.complexfloating)


class Constellation(abc.Constellation[T]):
    r"""
    General real or complex constellation. A *constellation* of *dimension* $N$ and *order* $M$ is defined by an ordered set $\\{ \mathbf{x}_i : i \in [0:M) \\}$ of $M$ distinct points in $\mathbb{R}^N$ or $\mathbb{C}^N$, called *symbols*. In this class, the constellation is represented by a matrix $\mathbf{X} \in \mathbb{R}^{M \times N}$ or $\mathbf{X} \in \mathbb{C}^{M \times N}$, where the $i$-th row of $\mathbf{X}$ corresponds to symbol $\mathbf{x}_i$. For more details, see <cite>SA15, Sec. 2.5.1</cite>.

    Parameters:
        matrix: The constellation matrix $\mathbf{X}$. Must be a 2D-array of shape $(M, N)$ with real or complex entries.

    Examples:
        The real constellation depicted in the figure below has $M = 5$ and $N = 2$.
        <figure markdown>
        ![Example for real constellation with M = 5 and N = 2](/fig/constellation_crux.svg)
        </figure>

            >>> const = komm.Constellation([[0, 4], [-2, 2], [2, 2], [1, 1], [0, -2]])
    """

    def __init__(self, matrix: npt.ArrayLike) -> None:
        dtype = complex if np.iscomplexobj(matrix) else float
        matrix = np.asarray(matrix, dtype=dtype)
        if matrix.ndim == 1:
            matrix = matrix.reshape(-1, 1)
        if matrix.ndim != 2:
            raise ValueError("'matrix' must be a 2D-array")
        self._matrix = matrix

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.matrix.tolist()})"

    @property
    def matrix(self) -> Array2D[T]:
        r"""
        Examples:
            >>> const = komm.Constellation([[0, 4], [-2, 2], [2, 2], [1, 1], [0, -2]])
            >>> const.matrix
            array([[ 0.,  4.],
                   [-2.,  2.],
                   [ 2.,  2.],
                   [ 1.,  1.],
                   [ 0., -2.]])
        """
        return self._matrix

    @property
    def order(self) -> int:
        r"""
        Examples:
            >>> const = komm.Constellation([[0, 4], [-2, 2], [2, 2], [1, 1], [0, -2]])
            >>> const.order
            5
        """
        return super().order

    @property
    def dimension(self) -> int:
        r"""
        Examples:
            >>> const = komm.Constellation([[0, 4], [-2, 2], [2, 2], [1, 1], [0, -2]])
            >>> const.dimension
            2
        """
        return super().dimension

    def mean(self, priors: npt.ArrayLike | None = None) -> Array1D[T]:
        r"""
        Examples:
            >>> const = komm.Constellation([[0, 4], [-2, 2], [2, 2], [1, 1], [0, -2]])
            >>> const.mean()
            array([0.2, 1.4])
        """
        return super().mean(priors)

    def mean_energy(self, priors: npt.ArrayLike | None = None) -> np.floating:
        r"""
        Examples:
            >>> const = komm.Constellation([[0, 4], [-2, 2], [2, 2], [1, 1], [0, -2]])
            >>> const.mean_energy()  # doctest: +FLOAT_CMP
            np.float64(7.6)
        """
        return super().mean_energy(priors)

    def minimum_distance(self) -> np.floating:
        r"""
        Examples:
            >>> const = komm.Constellation([[0, 4], [-2, 2], [2, 2], [1, 1], [0, -2]])
            >>> const.minimum_distance()  # doctest: +FLOAT_CMP
            np.float64(1.4142135623730951)
        """
        return super().minimum_distance()

    def indices_to_symbols(self, indices: npt.ArrayLike) -> npt.NDArray[T]:
        r"""
        Examples:
            >>> const = komm.Constellation([[0, 4], [-2, 2], [2, 2], [1, 1], [0, -2]])
            >>> const.indices_to_symbols([3, 0])
            array([1., 1., 0., 4.])
            >>> const.indices_to_symbols([[3, 0], [1, 2]])
            array([[ 1.,  1.,  0.,  4.],
                   [-2.,  2.,  2.,  2.]])
        """
        return super().indices_to_symbols(indices)

    def closest_indices(self, received: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> const = komm.Constellation([[0, 4], [-2, 2], [2, 2], [1, 1], [0, -2]])
            >>> const.closest_indices([0.3, 1.8, 0.0, 5.0])
            array([3, 0])
            >>> const.closest_indices([[0.3, 1.8], [0.0, 5.0]])
            array([[3],
                   [0]])
        """
        return super().closest_indices(received)

    def closest_symbols(self, received: npt.ArrayLike) -> npt.NDArray[T]:
        r"""
        Examples:
            >>> const = komm.Constellation([[0, 4], [-2, 2], [2, 2], [1, 1], [0, -2]])
            >>> const.closest_symbols([0.3, 1.8, 0.0, 5.0])
            array([1., 1., 0., 4.])
            >>> const.closest_symbols([[0.3, 1.8], [0.0, 5.0]])
            array([[1., 1.], [0., 4.]])
        """
        return super().closest_symbols(received)

    def posteriors(
        self,
        received: npt.ArrayLike,
        snr: float,
        priors: npt.ArrayLike | None = None,
    ) -> npt.NDArray[np.floating]:
        r"""
        Examples:
            >>> const = komm.Constellation([[0, 4], [-2, 2], [2, 2], [1, 1], [0, -2]])
            >>> const.posteriors([0.3, 1.8, 0.0, 5.0], snr=2.0).round(4)
            array([0.1565, 0.1408, 0.2649, 0.4253, 0.0125,
                   0.9092, 0.0387, 0.0387, 0.0135, 0.    ])
            >>> const.posteriors([[0.3, 1.8], [0.0, 5.0]], snr=2.0).round(4)
            array([[0.1565, 0.1408, 0.2649, 0.4253, 0.0125],
                   [0.9092, 0.0387, 0.0387, 0.0135, 0.    ]])
        """
        return super().posteriors(received, snr, priors)



from typing import Protocol, TypeVar, runtime_checkable

from typing_extensions import Self

from . import ring

T = TypeVar("T", bound="DomainElement")


@runtime_checkable
class DomainElement(ring.RingElement, Protocol):
    def __floordiv__(self: Self, other: Self) -> Self: ...
    def __mod__(self: Self, other: Self) -> Self: ...


def gcd(x: T, y: T) -> T:
    r"""
    Greatest common divisor. Computes the greatest common divisor (gcd) of two elements in a domain using the Euclidean algorithm.

    Parameters:
        x: First element
        y: Second element

    Returns:
        gcd: The greatest common divisor of `x` and `y`

    References:
        `Euclidean algorithm <https://en.wikipedia.org/wiki/Euclidean_algorithm>`_
    """
    zero = x.ambient.zero
    while y != zero:
        x, y = y, x % y
    return x


def xgcd(x: T, y: T) -> tuple[T, T, T]:
    r"""
    Extended gcd. Computes the greatest common divisor (gcd) and the Bézout coefficients of two elements in a domain using the extended Euclidean algorithm.

    Parameters:
        x: First element
        y: Second element

    Returns:
        gcd: The greatest common divisor of `x` and `y`
        s: The Bézout coefficient of `x`
        t: The Bézout coefficient of `y`

    References:
        `Extended Euclidean algorithm <https://en.wikipedia.org/wiki/Extended_Euclidean_algorithm>`_
    """
    zero: T = x.ambient.zero
    one: T = x.ambient.one
    if x == zero:
        return y, zero, one
    else:
        d, s, t = xgcd(y % x, x)
        return d, t - s * (y // x), s


def lcm(x: T, y: T) -> T:
    r"""
    Least common multiple. Computes the least common multiple (lcm) of two elements in a domain.

    Parameters:
        x: First element
        y: Second element

    Returns:
        lcm: The least common multiple of `x` and `y`

    References:
        `Least common multiple <https://en.wikipedia.org/wiki/Least_common_multiple>`_
    """
    zero = x.ambient.zero
    if x == zero or y == zero:
        return zero
    g = gcd(x, y)
    return (x // g) * y



from dataclasses import dataclass, field
from functools import cached_property
from typing import Literal

import numpy as np
import numpy.typing as npt

from .. import abc
from .._util import global_rng
from .._util.information_theory import binary_entropy
from .._util.validators import validate_pmf, validate_probability


@dataclass
class ZChannel(abc.DiscreteMemorylessChannel):
    r"""
    Z-channel. It is a [discrete memoryless channel](/ref/DiscreteMemorylessChannel) with input and output alphabets $\mathcal{X} = \mathcal{Y} = \\{ 0, 1 \\}$. The channel is characterized by a parameter $p$, called the *decay probability*. Bit $0$ is always received correctly, but bit $1$ turns into $0$ with probability $p$. Equivalently, the channel can be modeled as
    $$
        Y_n = A_n X_n,
    $$
    where $A_n$ are iid Bernoulli random variables with $\Pr[A_n = 0] = p$.

    Parameters:
        decay_probability: The channel decay probability $p$. Must satisfy $0 \leq p \leq 1$. The default value is `0.0`, which corresponds to a noiseless channel.
    """

    decay_probability: float = 0.0
    rng: np.random.Generator = field(default_factory=global_rng.get, repr=False)

    def __post_init__(self) -> None:
        validate_probability(self.decay_probability)

    @cached_property
    def input_cardinality(self) -> int:
        r"""
        For the Z-channel, it is given by $|\mathcal{X}| = 2$.
        """
        return 2

    @cached_property
    def output_cardinality(self) -> int:
        r"""
        For the Z-channel, it is given by $|\mathcal{Y}| = 2$.
        """
        return 2

    @cached_property
    def transition_matrix(self) -> npt.NDArray[np.floating]:
        r"""
        For the Z-channel, it is given by
        $$
            p_{Y \mid X} = \begin{bmatrix} 1 & 0 \\\\ p & 1-p \end{bmatrix}.
        $$

        Examples:
            >>> zc = komm.ZChannel(0.2)
            >>> zc.transition_matrix
            array([[1. , 0. ],
                   [0.2, 0.8]])
        """
        p = self.decay_probability
        return np.array([[1, 0], [p, 1 - p]])

    def mutual_information(
        self,
        input_pmf: npt.ArrayLike,
        base: float | Literal["e"] = 2.0,
    ) -> float:
        r"""
        For the Z-channel, it is given by
        $$
            \mathrm{I}(X ; Y) = \Hb \( \pi (1-p) \) - \pi \Hb(p),
        $$
        in bits, where $\pi = \Pr[X = 1]$, and $\Hb$ is the [binary entropy function](/ref/binary_entropy).

        Examples:
            >>> zc = komm.ZChannel(0.2)
            >>> zc.mutual_information([0.5, 0.5])  # doctest: +FLOAT_CMP
            np.float64(0.6099865470109874)
        """
        input_pmf = validate_pmf(input_pmf)
        p = self.decay_probability
        pi = input_pmf[1]
        return (binary_entropy(pi * (1 - p)) - pi * binary_entropy(p)) / np.log2(base)

    def capacity(self, base: float | Literal["e"] = 2.0) -> float:
        r"""
        For the Z-channel, it is given by
        $$
            C = \log_2 \( 1 + (1-p) p^{p / (1-p)} \),
        $$
        in bits.

        Examples:
            >>> zc = komm.ZChannel(0.2)
            >>> zc.capacity()  # doctest: +FLOAT_CMP
            np.float64(0.6182313659549211)
        """
        p = self.decay_probability
        if p == 1.0:
            return 0.0
        q = 1 - p
        return np.log2(1 + q * p ** (p / q)) / np.log2(base)

    def transmit(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> zc = komm.ZChannel(0.2)
            >>> zc.transmit([1, 1, 1, 0, 0, 0, 1, 0, 1, 0])
            array([1, 1, 1, 0, 0, 0, 1, 0, 0, 0])
        """
        p = self.decay_probability
        input = np.asarray(input)
        keep_pattern = self.rng.random(input.shape) >= p
        return (input * keep_pattern).astype(int)



from functools import cache

import numpy as np
import numpy.typing as npt

from .. import abc
from ..types import Array1D, Array2D


class PAMConstellation(abc.Constellation[np.floating]):
    r"""
    Pulse-amplitude modulation (PAM) constellation. It is a real one-dimensional [constellation](/ref/Constellation) in which the symbols are *uniformly arranged* in the real line and centered about the origin. For more details, see <cite>SA15, Sec. 2.5.1</cite>.

    Parameters:
        order: The order $M$ of the constellation.

        delta: The distance $\Delta$ between adjacent symbols. The default value is `2.0`.

    Examples:
        1. The $4$-PAM constellation with $\Delta = 2$ is depicted below.
            <figure markdown>
            ![4-PAM constellation.](/fig/constellation_pam_4.svg)
            </figure>

                >>> const = komm.PAMConstellation(4)

        1. The $7$-PAM constellation with $\Delta = 5$ is depicted below.
            <figure markdown>
            ![7-PAM constellation.](/fig/constellation_pam_7.svg)
            </figure>

                >>> const = komm.PAMConstellation(7, delta=5)
    """

    def __init__(self, order: int, delta: float = 2.0) -> None:
        self._order = order
        self._delta = np.float64(delta)

    def __repr__(self) -> str:
        args = ", ".join([
            f"order={self._order}",
            f"delta={self._delta}",
        ])
        return f"{self.__class__.__name__}({args})"

    @property
    @cache
    def matrix(self) -> Array2D[np.floating]:
        r"""
        Examples:
            >>> const = komm.PAMConstellation(4)
            >>> const.matrix
            array([[-3.],
                   [-1.],
                   [ 1.],
                   [ 3.]])
        """
        M, Δ = self._order, self._delta
        peak = (M - 1) * Δ / 2
        return np.linspace(-peak, peak, num=M, endpoint=True).reshape(-1, 1)

    @property
    def order(self) -> int:
        r"""
        Examples:
            >>> const = komm.PAMConstellation(4)
            >>> const.order
            4
        """
        return self._order

    @property
    def dimension(self) -> int:
        r"""
        For the PAM constellation, it is given by $N = 1$.

        Examples:
            >>> const = komm.PAMConstellation(4)
            >>> const.dimension
            1
        """
        return 1

    def mean(self, priors: npt.ArrayLike | None = None) -> Array1D[np.floating]:
        r"""
        For uniform priors, the mean of the PAM constellation is given by
        $$
            \mathbf{m} = 0.
        $$

        Examples:
            >>> const = komm.PAMConstellation(4)
            >>> const.mean()
            array([0.])
        """
        if priors is None:
            return np.zeros((1,), dtype=float)
        return super().mean(priors)

    def mean_energy(self, priors: npt.ArrayLike | None = None) -> np.floating:
        r"""
        For uniform priors, the mean energy of the PAM constellation is given by
        $$
            E = \frac{\Delta^2}{12}(M^2 - 1).
        $$

        Examples:
            >>> const = komm.PAMConstellation(4)
            >>> const.mean_energy()
            np.float64(5.0)
        """
        if priors is None:
            M, Δ = self._order, self._delta
            return np.float64(Δ**2 * (M**2 - 1) / 12)
        return super().mean_energy(priors)

    def minimum_distance(self) -> np.floating:
        r"""
        For the PAM constellation, the minimum distance is given by
        $$
            d_{\min} = \Delta.
        $$

        Examples:
            >>> const = komm.PAMConstellation(4)
            >>> const.minimum_distance()
            np.float64(2.0)
        """
        return self._delta

    def indices_to_symbols(self, indices: npt.ArrayLike) -> npt.NDArray[np.floating]:
        r"""
        Examples:
            >>> const = komm.PAMConstellation(4)
            >>> const.indices_to_symbols([3, 0])
            array([ 3., -3.])
            >>> const.indices_to_symbols([[3, 0], [1, 2]])
            array([[ 3., -3.],
                   [-1.,  1.]])
        """
        return super().indices_to_symbols(indices)

    def closest_indices(self, received: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> const = komm.PAMConstellation(4)
            >>> const.closest_indices([-0.8, 2.4])
            array([1, 3])
            >>> const.closest_indices([[-0.8, 2.4], [0.0, 10.0]])
            array([[1, 3],
                   [2, 3]])
        """
        M = self._order
        normalized = np.asarray(received, dtype=float) / self._delta
        indices = np.round(normalized + (M - 1) / 2).astype(int)
        indices = np.clip(indices, 0, M - 1)
        return indices

    def closest_symbols(self, received: npt.ArrayLike) -> npt.NDArray[np.floating]:
        r"""
        Examples:
            >>> const = komm.PAMConstellation(4)
            >>> const.closest_symbols([-0.8, 2.4])
            array([-1.,  3.])
            >>> const.closest_symbols([[-0.8, 2.4], [0.0, 10.0]])
            array([[-1.,  3.],
                   [ 1.,  3.]])
        """
        return super().closest_symbols(received)

    def posteriors(
        self,
        received: npt.ArrayLike,
        snr: float,
        priors: npt.ArrayLike | None = None,
    ) -> npt.NDArray[np.floating]:
        r"""
        Examples:
            >>> const = komm.PAMConstellation(4)
            >>> const.posteriors([-0.8, 2.4], snr=2.0).round(3)
            array([0.103, 0.7  , 0.195, 0.002, 0.   , 0.007, 0.343, 0.65 ])
            >>> const.posteriors([[-0.8, 2.4], [0.0, 10.0]], snr=2.0).round(3)
            array([[0.103, 0.7  , 0.195, 0.002, 0.   , 0.007, 0.343, 0.65 ],
                   [0.02 , 0.48 , 0.48 , 0.02 , 0.   , 0.   , 0.   , 1.   ]])
        """
        return super().posteriors(received, snr, priors)



from dataclasses import dataclass
from functools import cached_property

import numpy as np
import numpy.typing as npt

from .. import abc


@dataclass
class GaussianPulse(abc.Pulse):
    r"""
    Gaussian pulse. It is a [pulse](/ref/Pulse) with spectrum given by
    $$
        \hat{p}(f) = \frac{1}{\sqrt{2 \pi} \bar{B}} \mathrm{e}^{-\frac{1}{2} (f / \bar{B})^2},
    $$
    where the $\bar{B} = B / \sqrt{\ln 2}$, and $B$ is the _half-power bandwidth_ of the filter.

    The waveform of the Gaussian pulse is depicted below for $B = 0.5$, and for $B = 1$.

    <div class="centered" markdown>
      <span>
      ![Gaussian pulse with half-power bandwidth of 0.5.](/fig/pulse_gaussian_50.svg)
      </span>
      <span>
      ![Gaussian pulse with half-power bandwidth of 1.](/fig/pulse_gaussian_100.svg)
      </span>
    </div>

    Parameters:
        half_power_bandwidth: The half-power bandwidth $B$ of the pulse. The default value is `1.0`.
    """

    half_power_bandwidth: float = 1.0

    def waveform(self, t: npt.ArrayLike) -> npt.NDArray[np.floating]:
        r"""
        For the Gaussian pulse, it is given by
        $$
            p(t) = \mathrm{e}^{-\frac{1}{2} (2 \pi \bar{B} t)^2}.
        $$

        Examples:
            >>> pulse = komm.GaussianPulse(half_power_bandwidth=0.25)
            >>> pulse.waveform(
            ...     [-1.0, -0.75, -0.5, -0.25, 0.0, 0.25, 0.5, 0.75, 1.0],
            ... ).round(3)
            array([0.169, 0.367, 0.641, 0.895, 1.   , 0.895, 0.641, 0.367, 0.169])
        """
        t = np.asarray(t)
        b_bar = self.half_power_bandwidth / np.sqrt(np.log(2))
        return np.exp(-0.5 * (2 * np.pi * b_bar * t) ** 2)

    def spectrum(self, f: npt.ArrayLike) -> npt.NDArray[np.complexfloating]:
        r"""
        Examples:
            >>> pulse = komm.GaussianPulse(half_power_bandwidth=0.25)
            >>> np.abs(pulse.spectrum(
            ...     [-1.0, -0.75, -0.5, -0.25, 0.0, 0.25, 0.5, 0.75, 1.0],
            ... )).round(3)
            array([0.005, 0.059, 0.332, 0.939, 1.329, 0.939, 0.332, 0.059, 0.005])
        """
        f = np.asarray(f)
        b_bar = self.half_power_bandwidth / np.sqrt(np.log(2))
        spectrum = 1 / (np.sqrt(2 * np.pi) * b_bar) * np.exp(-0.5 * (f / b_bar) ** 2)
        return spectrum.astype(complex)

    def autocorrelation(self, tau: npt.ArrayLike) -> npt.NDArray[np.floating]:
        r"""
        For the Gaussian pulse, it is given by
        $$
            R(\tau) = \frac{1}{2 \sqrt{\pi} \bar{B}} \mathbb{e}^{-(\pi \bar{B} \tau)^2}.
        $$

        Examples:
            >>> pulse = komm.GaussianPulse(half_power_bandwidth=0.25)
            >>> pulse.autocorrelation(
            ...     [-1.0, -0.75, -0.5, -0.25, 0.0, 0.25, 0.5, 0.75, 1.0],
            ... ).round(3)
            array([0.386, 0.569, 0.752, 0.889, 0.939, 0.889, 0.752, 0.569, 0.386])
        """
        tau = np.asarray(tau)
        b_bar = self.half_power_bandwidth / np.sqrt(np.log(2))
        return (1 / (2 * np.sqrt(np.pi) * b_bar)) * np.exp(
            -((np.pi * b_bar * tau) ** 2)
        )

    def energy_density_spectrum(self, f: npt.ArrayLike) -> npt.NDArray[np.floating]:
        r"""
        For the Gaussian pulse, it is given by
        $$
            S(f) = \frac{1}{2 \pi \bar{B}^2} \mathrm{e}^{-(f / \bar{B})^2}.
        $$

        Examples:
            >>> pulse = komm.GaussianPulse(half_power_bandwidth=0.25)
            >>> pulse.energy_density_spectrum(
            ...     [-1.0, -0.75, -0.5, -0.25, 0.0, 0.25, 0.5, 0.75, 1.0],
            ... ).round(3)
            array([0.   , 0.003, 0.11 , 0.883, 1.765, 0.883, 0.11 , 0.003, 0.   ])
        """
        f = np.asarray(f)
        b_bar = self.half_power_bandwidth / np.sqrt(np.log(2))
        return (1 / (2 * np.pi * b_bar**2)) * np.exp(-((f / b_bar) ** 2))

    @cached_property
    def support(self) -> tuple[float, float]:
        r"""
        For the Gaussian pulse, the support is given by $(-\infty, \infty)$.

        Examples:
            >>> pulse = komm.GaussianPulse(half_power_bandwidth=0.25)
            >>> pulse.support
            (-inf, inf)
        """
        return (-np.inf, np.inf)

    def taps(
        self,
        samples_per_symbol: int,
        span: tuple[int, int] | None = None,
    ) -> npt.NDArray[np.floating]:
        r"""
        Examples:
            >>> pulse = komm.GaussianPulse(half_power_bandwidth=0.25)
            >>> pulse.taps(samples_per_symbol=4, span=(-1, 1)).round(3)
            array([0.169, 0.367, 0.641, 0.895, 1.   , 0.895, 0.641, 0.367, 0.169])
            >>> pulse.taps(samples_per_symbol=4, span=(-16, 16)).shape
            (129,)
        """
        return super().taps(samples_per_symbol, span)



from dataclasses import dataclass
from functools import cached_property

import numpy as np
import numpy.typing as npt

from .. import abc
from .util import rect, tri


@dataclass
class ManchesterPulse(abc.Pulse):
    r"""
    Manchester pulse. It is a [pulse](/ref/Pulse) with waveform given by
    $$
        p(t) =
        \begin{cases}
            -1, & 0 \leq t <  1/2, \\\\
            1, & 1/2 \leq t < 1, \\\\
            0, & \text{otherwise},
        \end{cases}
    $$

    The waveform of the Manchester pulse is depicted below.

    <figure markdown>
    ![Manchester pulse.](/fig/pulse_manchester.svg)
    </figure>

    <span class="doc-section-title">Parameters:</span>

    <span style="font-size: 90%; font-style: italic; color: gray; margin-left: 1em;">(No parameters)</span>
    """

    def waveform(self, t: npt.ArrayLike) -> npt.NDArray[np.floating]:
        r"""
        For the Manchester pulse, it is given by
        $$
            p(t) = -\rect\left(\frac{t - 1/4}{1/2}\right) + \rect\left(\frac{t - 3/4}{1/2}\right).
        $$

        Examples:
            >>> pulse = komm.ManchesterPulse()
            >>> pulse.waveform(
            ...     [-1.0, -0.75, -0.5, -0.25, 0.0, 0.25, 0.5, 0.75, 1.0],
            ... )
            array([ 0.,  0.,  0.,  0., -1., -1.,  1.,  1.,  0.])
        """
        t = np.asarray(t)
        return -rect(2 * (t - 0.25)) + rect(2 * (t - 0.75))

    def spectrum(self, f: npt.ArrayLike) -> npt.NDArray[np.complexfloating]:
        r"""
        For the Manchester pulse, it is given by
        $$
            \hat{p}(f) = \sinc \left( \frac{f}{2} \right) \, \sin \left( 2 \pi \frac{f}{4} \right) \mathrm{e}^{-\mathrm{j} 2 \pi (f/2\,+\,1/4)}.
        $$

        Examples:
            >>> pulse = komm.ManchesterPulse()
            >>> np.abs(pulse.spectrum(
            ...     [-1.0, -0.75, -0.5, -0.25, 0.0, 0.25, 0.5, 0.75, 1.0],
            ... )).round(3)
            array([0.637, 0.725, 0.637, 0.373, 0.   , 0.373, 0.637, 0.725, 0.637])
        """
        f = np.asarray(f)
        cexp = np.exp(-2j * np.pi * (f / 2 + 1 / 4))
        centered = np.sinc(f / 2) * np.sin(2 * np.pi * f / 4)
        return centered.astype(complex) * cexp

    def autocorrelation(self, tau: npt.ArrayLike) -> npt.NDArray[np.floating]:
        r"""
        For the Manchester pulse, it is given
        $$
            R(\tau) = \tri \left( \frac{\tau}{1/2} \right) - \frac{1}{2} \tri \left( \frac{\tau + 1/2}{1/2} \right) - \frac{1}{2} \tri \left( \frac{\tau - 1/2}{1/2} \right).
        $$

        Examples:
            >>> pulse = komm.ManchesterPulse()
            >>> pulse.autocorrelation(
            ...     [-1.0, -0.75, -0.5, -0.25, 0.0, 0.25, 0.5, 0.75, 1.0],
            ... )
            array([ 0.  , -0.25, -0.5 ,  0.25,  1.  ,  0.25, -0.5 , -0.25,  0.  ])
        """
        tau = np.asarray(tau)
        return tri(2 * tau) - 0.5 * tri(2 * (tau - 0.5)) - 0.5 * tri(2 * (tau + 0.5))

    def energy_density_spectrum(self, f: npt.ArrayLike) -> npt.NDArray[np.floating]:
        r"""
        For the Manchester pulse, it is given by
        $$
            S(f) = \sinc^2 \left( \frac{f}{2} \right) \, \sin^2 \left( 2 \pi \frac{f}{4} \right).
        $$

        Examples:
            >>> pulse = komm.ManchesterPulse()
            >>> pulse.energy_density_spectrum(
            ...     [-1.0, -0.75, -0.5, -0.25, 0.0, 0.25, 0.5, 0.75, 1.0],
            ... ).round(3)
            array([0.405, 0.525, 0.405, 0.139, 0.   , 0.139, 0.405, 0.525, 0.405])
        """
        f = np.asarray(f)
        return np.sinc(f / 2) ** 2 * np.sin(2 * np.pi * f / 4) ** 2

    @cached_property
    def support(self) -> tuple[float, float]:
        r"""
        For the Manchester pulse, the support is given by $[0, 1]$.

        Examples:
            >>> pulse = komm.ManchesterPulse()
            >>> pulse.support
            (0.0, 1.0)
        """
        return (0.0, 1.0)

    def taps(
        self,
        samples_per_symbol: int,
        span: tuple[int, int] | None = None,
    ) -> npt.NDArray[np.floating]:
        r"""
        Examples:
            >>> pulse = komm.ManchesterPulse()
            >>> pulse.taps(samples_per_symbol=4)  # Default span is [0, 1]
            array([-1., -1.,  1.,  1.,  0.])
            >>> pulse.taps(samples_per_symbol=4, span=(-1, 1))
            array([ 0., 0., 0., 0., -1., -1.,  1.,  1.,  0.])
        """
        return super().taps(samples_per_symbol, span)



from functools import cache
from math import comb

import numpy as np
import numpy.typing as npt
from typeguard import typechecked

from .._util.docs import mkdocstrings
from .BlockCode import BlockCode


@mkdocstrings(filters=["!.*"])
@typechecked
class SingleParityCheckCode(BlockCode):
    r"""
    Single parity-check code. For a given length $n \geq 1$, it is the [linear block code](/ref/BlockCode) whose codewords are obtained by extending $n - 1$ information bits with a single parity-check bit. The repetition code has the following parameters:

    - Length: $n$.
    - Dimension: $k = n - 1$.
    - Redundancy: $m = 1$.
    - Minimum distance: $d = 2$.

    Notes:
        - Its dual is the [repetition code](/ref/RepetitionCode).

    Parameters:
        n (int): The length $n$ of the code. Must be a positive integer.

    Examples:
        >>> code = komm.SingleParityCheckCode(5)
        >>> (code.length, code.dimension, code.redundancy)
        (5, 4, 1)
        >>> code.generator_matrix
        array([[1, 0, 0, 0, 1],
               [0, 1, 0, 0, 1],
               [0, 0, 1, 0, 1],
               [0, 0, 0, 1, 1]])
        >>> code.check_matrix
        array([[1, 1, 1, 1, 1]])
        >>> code.minimum_distance()
        2

        >>> code = komm.SingleParityCheckCode(16)
        >>> code.codeword_weight_distribution()
        array([    1,     0,   120,     0,  1820,     0,  8008,     0, 12870,
                   0,  8008,     0,  1820,     0,   120,     0,     1])
        >>> code.coset_leader_weight_distribution()
        array([1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
    """

    def __init__(self, n: int) -> None:
        if not n >= 1:
            raise ValueError("'n' must be a positive integer")
        self.n = n
        super().__init__(check_matrix=np.ones((1, n), dtype=int))

    def __repr__(self) -> str:
        args = f"n={self.n}"
        return f"{self.__class__.__name__}({args})"

    @cache
    def minimum_distance(self) -> int:
        return 2

    @cache
    def codeword_weight_distribution(self) -> npt.NDArray[np.integer]:
        n = self.n
        codeword_weight_distribution = np.zeros(n + 1, dtype=int)
        for w in range(0, n + 1, 2):
            codeword_weight_distribution[w] = comb(n, w)
        return codeword_weight_distribution



from dataclasses import dataclass
from math import ceil, log

import numpy as np
import numpy.typing as npt
from tqdm import tqdm

from komm._util.validators import validate_integer_range

from .util import Word, integer_to_symbols, symbols_to_integer


@dataclass
class LempelZivWelchCode:
    r"""
    Lempel–Ziv–Welch (LZW) code. It is a lossless data compression algorithm which is variation of the [Lempel–Ziv 78](/ref/LempelZiv78Code) algorithm. For more details, see <cite>Say06, Sec. 5.4.2</cite>.

    Note:
        Here, for simplicity, we assume that the source alphabet is $\mathcal{X} = [0 : |\mathcal{X}|)$ and the target alphabet is $\mathcal{Y} = [0 : |\mathcal{Y}|)$, where $|\mathcal{X}| \geq 2$ and $|\mathcal{Y}| \geq 2$ are called the *source cardinality* and *target cardinality*, respectively.

    Parameters:
        source_cardinality: The source cardinality $|\mathcal{X}|$. Must satisfy $|\mathcal{X}| \geq 2$.
        target_cardinality: The target cardinality $|\mathcal{Y}|$. Must satisfy $|\mathcal{Y}| \geq 2$. The default value is $2$ (binary).

    Examples:
        >>> lzw = komm.LempelZivWelchCode(2)  # Binary source, binary target
        >>> lzw = komm.LempelZivWelchCode(3, 4)  # Ternary source, quaternary target
    """

    source_cardinality: int
    target_cardinality: int = 2

    def __post_init__(self) -> None:
        if not self.source_cardinality >= 2:
            raise ValueError("'source_cardinality' must be at least 2")
        if not self.target_cardinality >= 2:
            raise ValueError("'target_cardinality' must be at least 2")

    def encode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Encodes a sequence of source symbols to a sequence of target symbols.

        Parameters:
            input: The sequence of source symbols to be encoded. Must be a 1D-array with elements in $\mathcal{X}$.

        Returns:
            output: The sequence of encoded target symbols. It is a 1D-array with elements in $\mathcal{Y}$.

        Examples:
            >>> lzw = komm.LempelZivWelchCode(2)
            >>> lzw.encode(np.zeros(15, dtype=int))
            array([0, 1, 0, 1, 1, 1, 0, 0, 1, 0, 1])

            >>> lzw = komm.LempelZivWelchCode(2, 8)
            >>> lzw.encode(np.zeros(15, dtype=int))
            array([0, 2, 3, 4, 5])
        """
        calX, calY = self.source_cardinality, self.target_cardinality
        input = validate_integer_range(input, high=calX)
        dictionary: dict[Word, int] = {(s,): s for s in range(calX)}
        output: list[int] = []

        word: Word = ()
        for symbol in tqdm(input, "Compressing LZW", delay=2.5):
            if word + (symbol,) in dictionary:
                word += (symbol,)
                continue
            k = ceil(log(len(dictionary), calY))
            pointer = dictionary[word]
            output.extend(integer_to_symbols(pointer, base=calY, width=k))
            dictionary[word + (symbol,)] = len(dictionary)
            word = (symbol,)

        if word:
            k = ceil(log(len(dictionary), calY))
            pointer = dictionary[word]
            output.extend(integer_to_symbols(pointer, base=calY, width=k))

        return np.array(output, dtype=int)

    def decode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Decodes a sequence of target symbols to a sequence of source symbols.

        Parameters:
            input: The sequence of target symbols to be decoded. Must be a 1D-array with elements in $\mathcal{Y}$. Also, the sequence must be a valid output of the `encode` method.

        Returns:
            output: The sequence of decoded source symbols. It is a 1D-array with elements in $\mathcal{X}$.

        Examples:
            >>> lzw = komm.LempelZivWelchCode(2)
            >>> lzw.decode([0, 1, 0, 1, 1, 1, 0, 0, 1, 0, 1])
            array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])

            >>> lzw = komm.LempelZivWelchCode(2, 8)
            >>> lzw.decode([0, 2, 3, 4, 5])
            array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
        """
        calX, calY = self.source_cardinality, self.target_cardinality
        input = np.asarray(input)

        if input.size == 0:
            return np.array([], dtype=int)

        dictionary: dict[int, Word] = {s: (s,) for s in range(calX)}
        output: list[int] = []

        k = ceil(log(calX, calY))
        pointer = symbols_to_integer(input[:k], base=calY)
        old = dictionary[pointer]
        output.extend(old)

        i = k
        pbar = tqdm(total=input.size, desc="Decompressing LZW", delay=2.5, initial=k)
        while True:
            k = ceil(log(len(dictionary) + 1, calY))
            if i + k > input.size:
                break
            pointer = symbols_to_integer(input[i : i + k], base=calY)
            word = dictionary.get(pointer, old + (old[0],))
            output.extend(word)
            dictionary[len(dictionary)] = old + (word[0],)
            old = word
            i += k
            pbar.update(k)
        pbar.close()

        return np.array(output, dtype=int)



from typing import Literal

import numpy as np
import numpy.typing as npt

from ..types import Array2D


def validate_log_base(value: float | str) -> float | Literal["e"]:
    if (isinstance(value, str) and value != "e") or (
        isinstance(value, float) and value <= 0.0
    ):
        raise ValueError("log base must be 'e' or a positive float")
    return value


def validate_probability(value: float) -> float:
    if not 0 <= value <= 1:
        raise ValueError("probability must be between 0 and 1")
    return value


def validate_pmf(
    value: npt.ArrayLike,
    joint: bool = False,
) -> npt.NDArray[np.floating]:
    value = np.asarray(value, dtype=float)
    if joint:
        if not (len(value.shape) > 0 and len(set(value.shape)) == 1):
            raise ValueError(
                f"pmf must have equal dimensions (got shape {value.shape})"
            )
    else:
        if not value.ndim == 1:
            raise ValueError("pmf must be a 1D-array")
    if not np.all(value >= 0.0):
        raise ValueError("pmf must be non-negative")
    if not np.isclose(value.sum(), 1.0):
        raise ValueError("pmf must sum to 1.0")
    return value


def validate_transition_matrix(
    value: npt.ArrayLike,
    square: bool = False,
) -> Array2D[np.floating]:
    value = np.asarray(value, dtype=float)
    if not value.ndim == 2:
        raise ValueError("transition matrix must be a 2D array")
    if not np.all(value >= 0.0):
        raise ValueError("transition matrix must be non-negative")
    if not np.allclose(value.sum(axis=1), 1.0):
        raise ValueError("rows of transition matrix must sum to 1.0")
    if square and value.shape[0] != value.shape[1]:
        raise ValueError(
            "transition matrix must be square (got shape "
            f"({value.shape[0]}, {value.shape[1]}))"
        )
    return value


def validate_integer_range(
    value: npt.ArrayLike,
    *,
    low: int = 0,
    high: int = 2,
) -> npt.NDArray[np.integer]:
    value = np.asarray(value, dtype=int)
    if not (np.all(value >= low) and np.all(value < high)):
        raise ValueError(f"input contains invalid entries (expected in [{low}:{high}))")
    return value



from collections.abc import Callable
from typing import Any, TypeVar

T = TypeVar("T")


def mkdocstrings(**options: Any) -> Callable[[type[T]], type[T]]:
    def decorator(cls: type[T]) -> type[T]:
        def _doc_options(cls: type[T]) -> dict[str, Any]:
            return options

        setattr(cls, "_doc_options", classmethod(_doc_options))
        return cls

    return decorator



from .BinarySequence import BinarySequence
from .sequences import barker_sequence


class BarkerSequence(BinarySequence):
    r"""
    Barker sequence. A Barker sequence is a [binary sequence](/ref/BinarySequence) with autocorrelation $R[\ell]$ satisfying $|R[\ell]| \leq 1$, for $\ell \neq 0$. The only known Barker sequences (up to negation and reversion) are shown in the table below. For more details, see [Wikipedia: Barker code](https://en.wikipedia.org/wiki/Barker_code).

    | Length $L$ | Barker sequence $b[n]$ |
    | :--------: | ---------------------- |
    | $2$        | $01$ and $00$          |
    | $3$        | $001$                  |
    | $4$        | $0010$ and $0001$      |
    | $5$        | $00010$                |
    | $7$        | $0001101$              |
    | $11$       | $00011101101$          |
    | $13$       | $0000011001010$        |

    Parameters:
        length: Length of the Barker sequence. Must be in the set $\\{ 2, 3, 4, 5, 7, 11, 13 \\}$.

    Examples:
        >>> barker = komm.BarkerSequence(length=13)
        >>> barker.polar_sequence
        array([ 1,  1,  1,  1,  1, -1, -1,  1,  1, -1,  1, -1,  1])
        >>> barker.autocorrelation()
        array([13,  0,  1,  0,  1,  0,  1,  0,  1,  0,  1,  0,  1])
    """

    def __init__(self, length: int) -> None:
        allowed_lengths = {2, 3, 4, 5, 7, 11, 13}
        if length not in allowed_lengths:
            raise ValueError(f"'length' must be in {allowed_lengths}")
        super().__init__(bit_sequence=barker_sequence(length))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(length={self.length})"



from functools import cache

import numpy as np
import numpy.typing as npt

from .. import abc
from ..types import Array1D, Array2D


class ASKConstellation(abc.Constellation[np.complexfloating]):
    r"""
    Amplitude-shift keying (ASK) constellation. It is a complex one-dimensional [constellation](/ref/Constellation) in which the symbols are *uniformly arranged* in a ray. More precisely, the $i$-th symbol is given by
    $$
        x_i = iA \exp(\mathrm{j} 2 \pi \phi), \quad i \in [0 : M),
    $$
    where $M$ is the *order*, $A$ is the *base amplitude*, and $\phi$ is the *phase offset* of the constellation.

    Parameters:
        order: The order $M$ of the constellation.

        base_amplitude: The base amplitude $A$ of the constellation. The default value is `1.0`.

        phase_offset: The phase offset $\phi$ of the constellation (in turns, not radians). The default value is `0.0`.

    Examples:
        1. The $4$-ASK constellation with base amplitude $A = 1$ and phase offset $\phi = 0$ is depicted below
            <figure markdown>
            ![4-ASK constellation.](/fig/constellation_ask_4.svg)
            </figure>

                >>> const = komm.ASKConstellation(4)

        1. The $4$-ASK constellation with base amplitude $A = 2\sqrt{2}$ and phase offset $\phi = 1/8$ is depicted below.
            <figure markdown>
            ![4-ASK constellation.](/fig/constellation_ask_4_turn.svg)
            </figure>

                >>> const = komm.ASKConstellation(
                ...     order=4,
                ...     base_amplitude=2 * np.sqrt(2),
                ...     phase_offset=1 / 8,
                ... )
    """

    def __init__(
        self, order: int, base_amplitude: float = 1.0, phase_offset: float = 0.0
    ) -> None:
        self._order = order
        self._base_amplitude = np.float64(base_amplitude)
        self._phase_offset = np.float64(phase_offset)

    def __repr__(self) -> str:
        args = ", ".join([
            f"order={self._order}",
            f"base_amplitude={self._base_amplitude}",
            f"phase_offset={self._phase_offset}",
        ])
        return f"{self.__class__.__name__}({args})"

    @property
    @cache
    def matrix(self) -> Array2D[np.complexfloating]:
        r"""
        Examples:
            >>> const = komm.ASKConstellation(4)
            >>> const.matrix
            array([[0.+0.j],
                   [1.+0.j],
                   [2.+0.j],
                   [3.+0.j]])
        """
        M, A, φ = self._order, self._base_amplitude, self._phase_offset
        i = np.arange(M).reshape(-1, 1)
        matrix = A * i * np.exp(2j * np.pi * φ)
        return matrix

    @property
    def order(self) -> int:
        r"""
        Examples:
            >>> const = komm.ASKConstellation(4)
            >>> const.order
            4
        """
        return self._order

    @property
    def dimension(self) -> int:
        r"""
        For the ASK constellation, it is given by $N = 1$.

        Examples:
            >>> const = komm.ASKConstellation(4)
            >>> const.dimension
            1
        """
        return 1

    def mean(self, priors: npt.ArrayLike | None = None) -> Array1D[np.complexfloating]:
        r"""
        For uniform priors, the mean of the ASK constellation is given by
        $$
            \mathbf{m} = \frac{A}{2} (M-1) \exp(\mathrm{j}\phi).
        $$

        Examples:
            >>> const = komm.ASKConstellation(4)
            >>> const.mean()
            array([1.5+0.j])
        """
        if priors is None:
            M, A, φ = self._order, self._base_amplitude, self._phase_offset
            m = 0.5 * A * (M - 1) * np.exp(2j * np.pi * φ)
            return np.array([m], dtype=complex)
        return super().mean(priors)

    def mean_energy(self, priors: npt.ArrayLike | None = None) -> np.floating:
        r"""
        For uniform priors, the mean energy of the ASK constellation is given by
        $$
            E = \frac{A^2}{6} (M - 1) (2M - 1).
        $$

        Examples:
            >>> const = komm.ASKConstellation(4)
            >>> const.mean_energy()
            np.float64(3.5)
        """
        if priors is None:
            M, A = self._order, self._base_amplitude
            return A**2 * (M - 1) * (2 * M - 1) / 6
        return super().mean_energy(priors)

    def minimum_distance(self) -> np.floating:
        r"""
        For the ASK constellation, the minimum distance is given by
        $$
            d_{\min} = A
        $$

        Examples:
            >>> const = komm.ASKConstellation(4)
            >>> const.minimum_distance()
            np.float64(1.0)
        """
        return self._base_amplitude

    def indices_to_symbols(
        self, indices: npt.ArrayLike
    ) -> npt.NDArray[np.complexfloating]:
        r"""
        Examples:
            >>> const = komm.ASKConstellation(4)
            >>> const.indices_to_symbols([3, 0])
            array([3.+0.j, 0.+0.j])
        """
        return super().indices_to_symbols(indices)

    def closest_indices(self, received: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> const = komm.ASKConstellation(4)
            >>> const.closest_indices([3.1 + 0.2j, 0.1 - 0.2j])
            array([3, 0])
        """
        M, A, φ = self._order, self._base_amplitude, self._phase_offset
        normalized = np.real(np.multiply(received, np.exp(-2j * np.pi * φ))) / A
        indices = np.round(normalized).astype(int)
        indices = np.clip(indices, 0, M - 1)
        return indices

    def closest_symbols(
        self, received: npt.ArrayLike
    ) -> npt.NDArray[np.complexfloating]:
        r"""
        Examples:
            >>> const = komm.ASKConstellation(4)
            >>> const.closest_symbols([3.1 + 0.2j, 0.1 - 0.2j])
            array([3.+0.j, 0.+0.j])
        """
        return super().closest_symbols(received)

    def posteriors(
        self,
        received: npt.ArrayLike,
        snr: float,
        priors: npt.ArrayLike | None = None,
    ) -> npt.NDArray[np.floating]:
        r"""
        Examples:
            >>> const = komm.ASKConstellation(4)
            >>> const.posteriors([3.1 + 0.2j, 0.1 - 0.2j], snr=5.0).round(3)
            array([0.   , 0.002, 0.152, 0.846, 0.755, 0.241, 0.004, 0.   ])
        """
        return super().posteriors(received, snr, priors)



from importlib import metadata as _metadata

__version__ = _metadata.version("komm")


# fmt: off
# isort: off

# Algebra
from ._algebra.BinaryPolynomial import BinaryPolynomial
from ._algebra.BinaryPolynomialFraction import BinaryPolynomialFraction
from ._algebra.FiniteBifield import FiniteBifield
from ._algebra.Integers import Integer

# Channels
from ._channels.AWGNChannel import AWGNChannel
from ._channels.BinaryErasureChannel import BinaryErasureChannel
from ._channels.BinarySymmetricChannel import BinarySymmetricChannel
from ._channels.DiscreteMemorylessChannel import DiscreteMemorylessChannel
from ._channels.ZChannel import ZChannel

# Constellations
from ._constellations.ASKConstellation import ASKConstellation
from ._constellations.APSKConstellation import APSKConstellation
from ._constellations.Constellation import Constellation
from ._constellations.PAMConstellation import PAMConstellation
from ._constellations.PSKConstellation import PSKConstellation
from ._constellations.QAMConstellation import QAMConstellation

# Error control - block codes
from ._error_control_block.BCHCode import BCHCode
from ._error_control_block.BlockCode import BlockCode
from ._error_control_block.CordaroWagnerCode import CordaroWagnerCode
from ._error_control_block.CyclicCode import CyclicCode
from ._error_control_block.GolayCode import GolayCode
from ._error_control_block.HammingCode import HammingCode
from ._error_control_block.Lexicode import Lexicode
from ._error_control_block.PolarCode import PolarCode
from ._error_control_block.ReedMullerCode import ReedMullerCode
from ._error_control_block.RepetitionCode import RepetitionCode
from ._error_control_block.SimplexCode import SimplexCode
from ._error_control_block.SingleParityCheckCode import SingleParityCheckCode
from ._error_control_block.SlepianArray import SlepianArray
from ._error_control_block.SystematicBlockCode import SystematicBlockCode

# Error control - checksum
from ._error_control_checksum.CyclicRedundancyCheck import CyclicRedundancyCheck

# Error control - convolutional
from ._error_control_convolutional.ConvolutionalCode import ConvolutionalCode
from ._error_control_convolutional.HighRateConvolutionalCode import HighRateConvolutionalCode
from ._error_control_convolutional.LowRateConvolutionalCode import LowRateConvolutionalCode
from ._error_control_convolutional.TerminatedConvolutionalCode import TerminatedConvolutionalCode
from ._error_control_convolutional.ViterbiStreamDecoder import ViterbiStreamDecoder

# Error control - decoders
from ._error_control_decoders.BCJRDecoder import BCJRDecoder
from ._error_control_decoders.BerlekampDecoder import BerlekampDecoder
from ._error_control_decoders.ExhaustiveSearchDecoder import ExhaustiveSearchDecoder
from ._error_control_decoders.ReedDecoder import ReedDecoder
from ._error_control_decoders.SCDecoder import SCDecoder
from ._error_control_decoders.SyndromeTableDecoder import SyndromeTableDecoder
from ._error_control_decoders.ViterbiDecoder import ViterbiDecoder
from ._error_control_decoders.WagnerDecoder import WagnerDecoder

# Finite state machines
from ._finite_state_machine.MealyMachine import MealyMachine
from ._finite_state_machine.MooreMachine import MooreMachine

# Integer coding
from ._integer_coding.FibonacciCode import FibonacciCode
from ._integer_coding.UnaryCode import UnaryCode

# Labelings
from ._labelings.Labeling import Labeling
from ._labelings.NaturalLabeling import NaturalLabeling
from ._labelings.ProductLabeling import ProductLabeling
from ._labelings.ReflectedLabeling import ReflectedLabeling
from ._labelings.ReflectedRectangularLabeling import ReflectedRectangularLabeling

# Lossless coding
from ._lossless_coding.FanoCode import FanoCode
from ._lossless_coding.FixedToVariableCode import FixedToVariableCode
from ._lossless_coding.HuffmanCode import HuffmanCode
from ._lossless_coding.LempelZiv77Code import LempelZiv77Code
from ._lossless_coding.LempelZiv78Code import LempelZiv78Code
from ._lossless_coding.LempelZivSSCode import LempelZivSSCode
from ._lossless_coding.LempelZivWelchCode import LempelZivWelchCode
from ._lossless_coding.ShannonCode import ShannonCode
from ._lossless_coding.TunstallCode import TunstallCode
from ._lossless_coding.VariableToFixedCode import VariableToFixedCode

# Pulses
from ._pulses.GaussianPulse import GaussianPulse
from ._pulses.ManchesterPulse import ManchesterPulse
from ._pulses.Pulse import Pulse
from ._pulses.RaisedCosinePulse import RaisedCosinePulse
from ._pulses.RectangularPulse import RectangularPulse
from ._pulses.RootRaisedCosinePulse import RootRaisedCosinePulse
from ._pulses.SincPulse import SincPulse

# Quantization
from ._quantization.LloydMaxQuantizer import LloydMaxQuantizer
from ._quantization.ScalarQuantizer import ScalarQuantizer
from ._quantization.UniformQuantizer import UniformQuantizer

# Sequences - binary
from ._sequences_binary.BarkerSequence import BarkerSequence
from ._sequences_binary.BinarySequence import BinarySequence
from ._sequences_binary.GoldSequence import GoldSequence
from ._sequences_binary.KasamiSequence import KasamiSequence
from ._sequences_binary.LFSRSequence import LFSRSequence
from ._sequences_binary.WalshHadamardSequence import WalshHadamardSequence

# Sequences - complex
from ._sequences_complex.ComplexSequence import ComplexSequence
from ._sequences_complex.ZadoffChuSequence import ZadoffChuSequence

# Sources
from ._sources.DiscreteMemorylessSource import DiscreteMemorylessSource
from ._sources.MarkovChain import MarkovChain

# Util
from ._util import global_rng
from ._util.bit_operations import bits_to_int, int_to_bits
from ._util.information_theory import binary_entropy, binary_entropy_inv, entropy, relative_entropy
from ._util.sequences import autocorrelation, cyclic_autocorrelation
from ._util.signal_processing import fourier_transform, sampling_rate_compress, sampling_rate_expand
from ._util.special_functions import boxplus, gaussian_q, gaussian_q_inv, marcum_q

# isort: on
# fmt: on


__all__ = [
    # _algebra
    "BinaryPolynomial",
    "BinaryPolynomialFraction",
    "FiniteBifield",
    "Integer",
    # _channels
    "AWGNChannel",
    "BinaryErasureChannel",
    "BinarySymmetricChannel",
    "DiscreteMemorylessChannel",
    "ZChannel",
    # _constellations
    "APSKConstellation",
    "ASKConstellation",
    "Constellation",
    "PAMConstellation",
    "PSKConstellation",
    "QAMConstellation",
    # _error_control_block
    "BCHCode",
    "BlockCode",
    "CordaroWagnerCode",
    "CyclicCode",
    "GolayCode",
    "HammingCode",
    "Lexicode",
    "PolarCode",
    "ReedMullerCode",
    "RepetitionCode",
    "SimplexCode",
    "SingleParityCheckCode",
    "SlepianArray",
    "SystematicBlockCode",
    # _error_control_checksum
    "CyclicRedundancyCheck",
    # _error_control_convolutional
    "ConvolutionalCode",
    "HighRateConvolutionalCode",
    "LowRateConvolutionalCode",
    "TerminatedConvolutionalCode",
    "ViterbiStreamDecoder",
    # _error_control_decoders
    "BCJRDecoder",
    "BerlekampDecoder",
    "ExhaustiveSearchDecoder",
    "ReedDecoder",
    "SCDecoder",
    "SyndromeTableDecoder",
    "ViterbiDecoder",
    "WagnerDecoder",
    # _finite_state_machine
    "MealyMachine",
    "MooreMachine",
    # _integer_coding
    "FibonacciCode",
    "UnaryCode",
    # _labelings
    "Labeling",
    "NaturalLabeling",
    "ProductLabeling",
    "ReflectedLabeling",
    "ReflectedRectangularLabeling",
    # _lossless_coding
    "FanoCode",
    "FixedToVariableCode",
    "HuffmanCode",
    "LempelZiv77Code",
    "LempelZiv78Code",
    "LempelZivSSCode",
    "LempelZivWelchCode",
    "ShannonCode",
    "TunstallCode",
    "VariableToFixedCode",
    # _pulses
    "GaussianPulse",
    "ManchesterPulse",
    "Pulse",
    "RaisedCosinePulse",
    "RectangularPulse",
    "RootRaisedCosinePulse",
    "SincPulse",
    # _quantization
    "LloydMaxQuantizer",
    "ScalarQuantizer",
    "UniformQuantizer",
    # _sequences_binary
    "BarkerSequence",
    "BinarySequence",
    "GoldSequence",
    "KasamiSequence",
    "LFSRSequence",
    "WalshHadamardSequence",
    # _sequences_complex
    "ComplexSequence",
    "ZadoffChuSequence",
    # _sources
    "DiscreteMemorylessSource",
    "MarkovChain",
    # _util
    "global_rng",
    "bits_to_int",
    "int_to_bits",
    "autocorrelation",
    "cyclic_autocorrelation",
    "binary_entropy",
    "binary_entropy_inv",
    "entropy",
    "relative_entropy",
    "fourier_transform",
    "sampling_rate_compress",
    "sampling_rate_expand",
    "boxplus",
    "gaussian_q",
    "gaussian_q_inv",
    "marcum_q",
]



from functools import cache
from itertools import product
from math import prod
from typing import TypeVar

import numpy as np
import numpy.typing as npt

from .. import abc
from .Labeling import Labeling

T = TypeVar("T", bound=np.generic)


class ProductLabeling(abc.Labeling):
    r"""
    Cartesian product of [labelings](/ref/Labeling).

    Parameters:
        *labelings: The labelings to be combined. At least one labeling is required.

        repeat: Number of times to repeat the full sequence of labelings. Must be a positive integer. Has the same semantics as `itertools.product`. The default value is `1`.

    Examples:
        >>> labeling1 = komm.Labeling([[0, 0], [1, 1], [1, 0], [0, 1]])
        >>> labeling2 = komm.Labeling([[1], [0]])
        >>> labeling = komm.ProductLabeling(labeling1, labeling2)
        >>> labeling.matrix
        array([[0, 0, 1],
               [0, 0, 0],
               [1, 1, 1],
               [1, 1, 0],
               [1, 0, 1],
               [1, 0, 0],
               [0, 1, 1],
               [0, 1, 0]])


        >>> labeling = komm.ProductLabeling(
        ...    komm.Labeling([[1, 0], [1, 1], [0, 1], [0, 0]]),
        ...    repeat=2,
        ... )
        >>> labeling.matrix
        array([[1, 0, 1, 0],
               [1, 0, 1, 1],
               [1, 0, 0, 1],
               [1, 0, 0, 0],
               [1, 1, 1, 0],
               [1, 1, 1, 1],
               [1, 1, 0, 1],
               [1, 1, 0, 0],
               [0, 1, 1, 0],
               [0, 1, 1, 1],
               [0, 1, 0, 1],
               [0, 1, 0, 0],
               [0, 0, 1, 0],
               [0, 0, 1, 1],
               [0, 0, 0, 1],
               [0, 0, 0, 0]])
    """

    def __init__(self, *labelings: abc.Labeling, repeat: int = 1) -> None:
        if len(labelings) < 1:
            raise ValueError("at least one labeling is required")
        if repeat < 1:
            raise ValueError("'repeat' must be at least 1")
        self._labelings = labelings * repeat

    @classmethod
    def from_matrices(
        cls,
        *matrices: npt.ArrayLike,
        repeat: int = 1,
    ) -> "ProductLabeling":
        r"""
        Constructs a product labeling from labeling matrices.

        Parameters:
            *matrices: The labeling matrices. At least one matrix is required. See [labeling documentation](/ref/Labeling).

            repeat: Number of times to repeat the full sequence of matrices. Must be a positive integer. Has the same semantics as `itertools.product`. The default value is `1`.

        Examples:
            >>> labeling = komm.ProductLabeling.from_matrices(
            ...     [[0, 0], [1, 1], [1, 0], [0, 1]],
            ...     [[1], [0]],
            ... )
            >>> labeling.matrix
            array([[0, 0, 1],
                   [0, 0, 0],
                   [1, 1, 1],
                   [1, 1, 0],
                   [1, 0, 1],
                   [1, 0, 0],
                   [0, 1, 1],
                   [0, 1, 0]])

            >>> labeling = komm.ProductLabeling.from_matrices(
            ...     [[1, 0], [1, 1], [0, 1], [0, 0]],
            ...     repeat=2,
            ... )
            >>> labeling.matrix
            array([[1, 0, 1, 0],
                   [1, 0, 1, 1],
                   [1, 0, 0, 1],
                   [1, 0, 0, 0],
                   [1, 1, 1, 0],
                   [1, 1, 1, 1],
                   [1, 1, 0, 1],
                   [1, 1, 0, 0],
                   [0, 1, 1, 0],
                   [0, 1, 1, 1],
                   [0, 1, 0, 1],
                   [0, 1, 0, 0],
                   [0, 0, 1, 0],
                   [0, 0, 1, 1],
                   [0, 0, 0, 1],
                   [0, 0, 0, 0]])
        """
        labelings = tuple(Labeling(m) for m in matrices)
        return cls(*labelings, repeat=repeat)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._labelings})"

    @property
    @cache
    def matrix(self) -> npt.NDArray[np.integer]:
        matrices = [lab.matrix for lab in self._labelings]
        rows = [np.hstack(comb) for comb in product(*matrices)]
        return np.vstack(rows)

    @property
    def num_bits(self) -> int:
        r"""
        Examples:
            >>> labeling = komm.ProductLabeling.from_matrices(
            ...     [[0, 0], [1, 1], [1, 0], [0, 1]],
            ...     [[1], [0]],
            ... )
            >>> labeling.num_bits
            3
        """
        return sum(lab.num_bits for lab in self._labelings)

    @property
    def cardinality(self) -> int:
        r"""
        Examples:
            >>> labeling = komm.ProductLabeling.from_matrices(
            ...     [[0, 0], [1, 1], [1, 0], [0, 1]],
            ...     [[1], [0]],
            ... )
            >>> labeling.cardinality
            8
        """
        return prod(lab.cardinality for lab in self._labelings)

    @property
    @cache
    def inverse_mapping(self) -> dict[tuple[int, ...], int]:
        r"""
        Examples:
            >>> labeling = komm.ProductLabeling.from_matrices(
            ...     [[0, 0], [1, 1], [1, 0], [0, 1]],
            ...     [[1], [0]],
            ... )
            >>> labeling.inverse_mapping
            {(0, 0, 1): 0,
             (0, 0, 0): 1,
             (1, 1, 1): 2,
             (1, 1, 0): 3,
             (1, 0, 1): 4,
             (1, 0, 0): 5,
             (0, 1, 1): 6,
             (0, 1, 0): 7}
        """
        return super().inverse_mapping

    def indices_to_bits(self, indices: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> labeling = komm.ProductLabeling.from_matrices(
            ...     [[0, 0], [1, 1], [1, 0], [0, 1]],
            ...     [[1], [0]],
            ... )
            >>> labeling.indices_to_bits([2, 0])
            array([1, 1, 1, 0, 0, 1])
        """
        return super().indices_to_bits(indices)

    def bits_to_indices(self, bits: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> labeling = komm.ProductLabeling.from_matrices(
            ...     [[0, 0], [1, 1], [1, 0], [0, 1]],
            ...     [[1], [0]],
            ... )
            >>> labeling.bits_to_indices([1, 1, 1, 0, 0, 1])
            array([2, 0])
        """
        return super().bits_to_indices(bits)

    def marginalize(self, metrics: npt.ArrayLike) -> npt.NDArray[np.floating]:
        r""" """
        return super().marginalize(metrics)



from dataclasses import dataclass
from functools import cached_property

import numpy as np
import numpy.typing as npt

from .. import abc
from .util import rect


@dataclass
class SincPulse(abc.Pulse):
    r"""
    Sinc pulse. It is a [pulse](/ref/Pulse) with waveform given by
    $$
        p(t) = \frac{\sin(\pi t)}{\pi t} = \sinc(t).
    $$

    The waveform of the sinc pulse is depicted below.

    <figure markdown>
    ![Sinc pulse.](/fig/pulse_sinc.svg)
    </figure>

    <span class="doc-section-title">Parameters:</span>

    <span style="font-size: 90%; font-style: italic; color: gray; margin-left: 1em;">(No parameters)</span>
    """

    def waveform(self, t: npt.ArrayLike) -> npt.NDArray[np.floating]:
        r"""
        For the sinc pulse, it is given by
        $$
            p(t) = \sinc(t).
        $$

        Examples:
            >>> pulse = komm.SincPulse()
            >>> pulse.waveform(
            ...     [-1.0, -0.75, -0.5, -0.25, 0.0, 0.25, 0.5, 0.75, 1.0],
            ... ).round(3)
            array([0.   , 0.3  , 0.637, 0.9  , 1.   , 0.9  , 0.637, 0.3  , 0.   ])
        """
        t = np.asarray(t)
        return np.sinc(t)

    def spectrum(self, f: npt.ArrayLike) -> npt.NDArray[np.complexfloating]:
        r"""
        For the sinc pulse, it is given by
        $$
            \hat{p}(f) = \rect(f).
        $$

        Examples:
            >>> pulse = komm.SincPulse()
            >>> np.abs(pulse.spectrum(
            ...     [-1.0, -0.75, -0.5, -0.25, 0.0, 0.25, 0.5, 0.75, 1.0],
            ... ))
            array([0., 0., 1., 1., 1., 1., 0., 0., 0.])
        """
        f = np.asarray(f)
        return rect(f).astype(complex)

    def autocorrelation(self, tau: npt.ArrayLike) -> npt.NDArray[np.floating]:
        r"""
        For the sinc pulse, it is given by
        $$
            R(\tau) = \sinc(\tau).
        $$

        Examples:
            >>> pulse = komm.SincPulse()
            >>> pulse.autocorrelation(
            ...     [-1.0, -0.75, -0.5, -0.25, 0.0, 0.25, 0.5, 0.75, 1.0],
            ... ).round(3)
            array([0.   , 0.3  , 0.637, 0.9  , 1.   , 0.9  , 0.637, 0.3  , 0.   ])
        """
        tau = np.asarray(tau)
        return np.sinc(tau)

    def energy_density_spectrum(self, f: npt.ArrayLike) -> npt.NDArray[np.floating]:
        r"""
        For the sinc pulse, it is given by
        $$
            S(f) = \rect(f).
        $$

        Examples:
            >>> pulse = komm.SincPulse()
            >>> pulse.energy_density_spectrum(
            ...     [-1.0, -0.75, -0.5, -0.25, 0.0, 0.25, 0.5, 0.75, 1.0],
            ... )
            array([0., 0., 1., 1., 1., 1., 0., 0., 0.])
        """
        f = np.asarray(f)
        return rect(f)

    @cached_property
    def support(self) -> tuple[float, float]:
        r"""
        For the sinc pulse, the support is given by $(-\infty, \infty)$.

        Examples:
            >>> pulse = komm.SincPulse()
            >>> pulse.support
            (-inf, inf)
        """
        return (-np.inf, np.inf)

    def taps(
        self,
        samples_per_symbol: int,
        span: tuple[int, int] | None = None,
    ) -> npt.NDArray[np.floating]:
        r"""
        Examples:
            >>> pulse = komm.SincPulse()
            >>> pulse.taps(samples_per_symbol=4, span=(-1, 1)).round(3)
            array([0.   , 0.3  , 0.637, 0.9  , 1.   , 0.9  , 0.637, 0.3  , 0.   ])
            >>> pulse.taps(samples_per_symbol=4, span=(-16, 16)).shape
            (129,)
        """
        return super().taps(samples_per_symbol, span)



from typing_extensions import Self

from .._algebra.BinaryPolynomial import BinaryPolynomial, default_primitive_polynomial
from .BinarySequence import BinarySequence
from .sequences import lfsr_sequence


class LFSRSequence(BinarySequence):
    r"""
    Linear-feedback shift register (LFSR) sequence. It is a [binary sequence](/ref/BinarySequence) obtained from the output of a LFSR. The LFSR feedback taps are specified as a [binary polynomial](/ref/BinaryPolynomial) $p(X)$ of degree $n$, called the *feedback polynomial*. More specifically: if bit $i$ of the LFSR is tapped, for $i \in [1 : n]$, then the coefficient of $X^i$ in $p(X)$ is $1$; otherwise, it is $0$; moreover, the coefficient of $X^0$ in $p(X)$ is always $1$. For example, the feedback polynomial corresponding to the LFSR in the figure below is $p(X) = X^5 + X^2 + 1$, whose integer representation is `0b100101`.

    <figure markdown>
      ![Linear-feedback shift register example.](/fig/lfsr_5_2.svg)
    </figure>

    The start state of the machine is specified by the so called *start state polynomial*. More specifically, the coefficient of $X^i$ in the start state polynomial is equal to the initial value of bit $i$ of the LFSR. For more details, see [Wikipedia: Linear-feedback shift register](https://en.wikipedia.org/wiki/Linear-feedback_shift_register) and [Wikipedia: Maximum-length sequence](https://en.wikipedia.org/wiki/Maximum_length_sequence).

    The default constructor of this takes the following parameters:

    Parameters:
        feedback_polynomial: The feedback polynomial $p(X)$ of the LFSR, specified either as a [binary polynomial](/ref/BinaryPolynomial) or as an integer to be converted to the former.

        start_state_polynomial: The start state polynomial of the LFSR, specified either as a [binary polynomial](/ref/BinaryPolynomial) or as an integer to be converted to the former. The default value is `0b1`.

    Examples:
        >>> lfsr = komm.LFSRSequence(feedback_polynomial=0b100101)
        >>> lfsr.bit_sequence
        array([0, 0, 0, 0, 1, 0, 1, 0, 1, 1, 1, 0, 1, 1, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1])
        >>> lfsr.cyclic_autocorrelation()
        array([31, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1])

    <h2>Maximum-length sequences</h2>

    If the feedback polynomial $p(X)$ is primitive, then the corresponding LFSR sequence will be a *maximum-length sequence*. Such sequences have the following cyclic autocorrelation:
    $$
        \tilde{R}[\ell] =
        \begin{cases}
            L, & \ell = 0, \, \pm L, \, \pm 2L, \ldots, \\\\
            -1, & \text{otherwise},
        \end{cases}
    $$
    where $L$ is the length of the sequence. See the class method [`maximum_length_sequence`](./#maximum_length_sequence) for a convenient way to construct a maximum-length sequence.
    """

    def __init__(
        self,
        feedback_polynomial: BinaryPolynomial | int,
        start_state_polynomial: BinaryPolynomial | int = 0b1,
    ) -> None:
        self.feedback_polynomial = BinaryPolynomial(feedback_polynomial)
        self.start_state_polynomial = BinaryPolynomial(start_state_polynomial)
        super().__init__(
            bit_sequence=lfsr_sequence(
                self.feedback_polynomial, self.start_state_polynomial
            )
        )

    @classmethod
    def maximum_length_sequence(
        cls, degree: int, start_state_polynomial: BinaryPolynomial | int = 0b1
    ) -> Self:
        r"""
        Constructs a maximum-length sequences of a given degree. The feedback polynomial $p(X)$ is chosen from [the list of default primitive polynomials](/res/primitive-polynomials).

        Parameters:
            degree: The degree $n$ of the maximum-length-sequence. Only degrees in the range $[1 : 24]$ are implemented.

            start_state_polynomial: See the corresponding parameter of the default constructor.

        Examples:
            >>> komm.LFSRSequence.maximum_length_sequence(degree=5)
            LFSRSequence(feedback_polynomial=0b100101, start_state_polynomial=0b1)
        """
        return cls(
            feedback_polynomial=default_primitive_polynomial(degree),
            start_state_polynomial=start_state_polynomial,
        )

    def __repr__(self) -> str:
        args = (
            f"feedback_polynomial={self.feedback_polynomial}, "
            f"start_state_polynomial={self.start_state_polynomial}"
        )
        return f"{self.__class__.__name__}({args})"



import operator
from functools import reduce
from typing import SupportsInt, TypeVar

import numpy as np
import numpy.typing as npt
from typing_extensions import Self

from .._util.bit_operations import bits_to_int, int_to_bits
from . import domain, ring
from .Integers import prime_factors

T = TypeVar("T", bound=ring.RingElement)


class BinaryPolynomial:
    r"""
    Binary polynomial. A *binary polynomial* is a polynomial whose coefficients are elements in the finite field $\mathbb{F}_2 = \\{ 0, 1 \\}$.

    The default constructor of the class expects the following:

    Parameters:
        value: An integer whose binary digits represent the coefficients of the polynomial—the leftmost bit standing for the highest degree term. For example, the binary polynomial $X^4 + X^3 + X$ is represented by the integer `0b11010` = `0o32` = `26`.

    Examples:
        >>> komm.BinaryPolynomial(0b11010)  # X^4 + X^3 + X
        BinaryPolynomial(0b11010)

    See also the class methods [`from_coefficients`](./#from_coefficients) and [`from_exponents`](./#from_exponents) for alternative ways to construct a binary polynomial.

    <h2>Algebraic structure</h2>

    The binary polynomials form an *Euclidean domain*. The following operations are supported: addition (`+`), subtraction (`-`), multiplication (`*`), euclidean division (`//`), modulo (`%`), and exponentiation (`**`).

    Examples:
        >>> poly1 = komm.BinaryPolynomial(0b10111)  # X^4 + X^2 + X + 1
        >>> poly2 = komm.BinaryPolynomial(0b101)  # X^2 + 1
        >>> poly1 + poly2  # X^4 + X
        BinaryPolynomial(0b10010)
        >>> poly1 - poly2  # X^4 + X
        BinaryPolynomial(0b10010)
        >>> poly1 * poly2  # X^6 + X^3 + X + 1
        BinaryPolynomial(0b1001011)
        >>> poly1 // poly2  # X^2
        BinaryPolynomial(0b100)
        >>> poly1 % poly2  # X + 1
        BinaryPolynomial(0b11)
        >>> poly1 ** 2  # X^8 + X^4 + X^2 + 1
        BinaryPolynomial(0b100010101)
    """

    def __init__(self, value: SupportsInt) -> None:
        self.value = int(value)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.value:#b})"

    def __str__(self) -> str:
        return bin(self.value)

    def __int__(self) -> int:
        return self.value

    def __hash__(self) -> int:
        return self.value

    @property
    def ambient(self):
        return BinaryPolynomials()

    @classmethod
    def from_coefficients(cls, coefficients: npt.ArrayLike) -> Self:
        r"""
        Constructs a binary polynomial from its coefficients.

        Parameters:
            coefficients: The coefficients of the binary polynomial—the $i$-th element of the array standing for the coefficient of $X^i$. For example, `[0, 1, 0, 1, 1]` represents the binary polynomial $X^4 + X^3 + X$.

        Examples:
            >>> komm.BinaryPolynomial.from_coefficients([0, 1, 0, 1, 1])  # X^4 + X^3 + X
            BinaryPolynomial(0b11010)
        """
        return cls(bits_to_int(coefficients))

    @classmethod
    def from_exponents(cls, exponents: npt.ArrayLike) -> Self:
        r"""
        Constructs a binary polynomial from its exponents.

        Parameters:
            exponents: The exponents of the nonzero terms of the binary polynomial. For example, `[1, 3, 4]` represents the binary polynomial $X^4 + X^3 + X$.

        Examples:
            >>> komm.BinaryPolynomial.from_exponents([1, 3, 4])  # X^4 + X^3 + X
            BinaryPolynomial(0b11010)
        """
        return cls(bits_to_int(np.bincount(exponents)))

    def __eq__(self, other: object) -> bool:
        if isinstance(other, int):
            return self.value == other
        if isinstance(other, self.__class__):
            return self.value == other.value
        return NotImplemented

    def __lshift__(self, n: int) -> Self:
        return self.__class__(self.value << n)

    def __rshift__(self, n: int) -> Self:
        return self.__class__(self.value >> n)

    def __add__(self, other: Self) -> Self:
        return self.__class__(self.value ^ other.value)

    def __sub__(self, other: Self) -> Self:
        return self.__class__(self.value ^ other.value)

    def __neg__(self) -> Self:
        return self

    def __mul__(self, other: Self) -> Self:
        value = 0
        a, b = self.value, other.value
        while b:
            if b & 1:
                value ^= a
            a <<= 1
            b >>= 1
        return self.__class__(value)

    def __rmul__(self, other: int) -> Self:
        if other % 2 == 0:
            return self.__class__(0)
        else:
            return self

    def __pow__(self, exponent: int) -> Self:
        return ring.power(self, exponent)

    def __divmod__(self, den: Self) -> tuple[Self, Self]:
        if den.value == 0:
            raise ZeroDivisionError("division by zero polynomial")
        div, mod = 0, self.value
        d = mod.bit_length() - den.value.bit_length()
        while d >= 0:
            div ^= 1 << d
            mod ^= den.value << d
            d = mod.bit_length() - den.value.bit_length()
        return self.__class__(div), self.__class__(mod)

    def __floordiv__(self, other: Self) -> Self:
        return self.__divmod__(other)[0]

    def __mod__(self, other: Self) -> Self:
        return self.__divmod__(other)[1]

    @property
    def degree(self) -> int:
        r"""
        The degree of the binary polynomial.

        Examples:
            >>> poly = komm.BinaryPolynomial(0b11010)  # X^4 + X^3 + X
            >>> poly.degree
            4
        """
        return self.value.bit_length() - 1

    def coefficients(self, width: int | None = None) -> npt.NDArray[np.integer]:
        r"""
        Returns the coefficients of the binary polynomial.

        Parameters:
            width: If this parameter is specified, the output will be filled with zeros on the right so that the its length will be the specified value.

        Returns:
            coefficients: Coefficients of the binary polynomial. The $i$-th element of the array stands for the coefficient of $X^i$.

        Examples:
            >>> poly = komm.BinaryPolynomial(0b11010)  # X^4 + X^3 + X
            >>> poly.coefficients()
            array([0, 1, 0, 1, 1])
            >>> poly.coefficients(width=8)
            array([0, 1, 0, 1, 1, 0, 0, 0])
        """
        return int_to_bits(self.value, width=width or max(self.degree + 1, 1))

    def exponents(self) -> npt.NDArray[np.integer]:
        r"""
        Returns the exponents of the binary polynomial.

        Returns:
            exponents: Exponents of the nonzero terms of the binary polynomial. The exponents are returned in ascending order.

        Examples:
            >>> poly = komm.BinaryPolynomial(0b11010)  # X^4 + X^3 + X
            >>> poly.exponents()
            array([1, 3, 4])
        """
        return np.flatnonzero(self.coefficients())

    def reciprocal(self, width: int | None = None) -> Self:
        r"""
        Returns the reciprocal (reflexion) of the binary polynomial. The reciprocal of a binary polynomial is the polynomial with the coefficients in reversed order, that is, the coefficient of $X^i$ becomes the coefficient of $X^{n-i}$, where $n$ is the degree of the polynomial.

        Returns:
            result (BinaryPolynomial): The binary polynomial with the reversed coefficients.

        Examples:
            >>> poly = komm.BinaryPolynomial(0b11010)  # X^4 + X^3 + X
            >>> poly.reciprocal()  # X^3 + X + 1
            BinaryPolynomial(0b1011)
        """
        return self.__class__(bits_to_int(self.coefficients(width)[::-1]))

    def evaluate(self, point: T) -> T:
        r"""
        Evaluates the binary polynomial at a given point. Uses Horner's method.

        Parameters:
            point (RingElement): The point at which the polynomial is evaluated. It must be an element of a ring in which multiplication by integers is defined.

        Returns:
            result (RingElement): The result of evaluating the binary polynomial at `point`. It has the same type as `point`.

        Examples:
            >>> poly = komm.BinaryPolynomial(0b11010)  # X^4 + X^3 + X
            >>> poly.evaluate(komm.Integer(7))  # same as 7**4 + 7**3 + 7
            Integer(2751)
        """
        return ring.binary_horner(self.coefficients(), point)

    def is_irreducible(self) -> bool:
        r"""
        Checks whether the binary polynomial is irreducible. A binary polynomial is *irreducible* if it is not divisible by any other nonconstant binary polynomial of smaller degree.

        This method checks the irreducibility using [Rabin's irreducibility test](https://en.wikipedia.org/wiki/Factorization_of_polynomials_over_finite_fields#Rabin's_test_of_irreducibility).

        Returns:
            result: `True`, if the binary polynomial is irreducible; `False`, otherwise.

        Examples:
            >>> komm.BinaryPolynomial(0b11011).is_irreducible()
            False
            >>> komm.BinaryPolynomial(0b11111).is_irreducible()
            True
            >>> komm.BinaryPolynomial(0b10011).is_irreducible()
            True
        """
        degree = self.degree
        if degree <= 0:  # p(X) = 0 or p(X) = 1
            return False
        X = BinaryPolynomial(0b10)  # The polynomial X
        for q in set(prime_factors(degree)):
            h = power_mod(X, 2 ** (degree // q), self) + X % self
            if BinaryPolynomial.gcd(self, h) != BinaryPolynomial(1):
                return False
        h = power_mod(X, 2**degree, self) + X % self
        return h == BinaryPolynomial(0)

    def is_primitive(self) -> bool:
        r"""
        Checks whether the binary polynomial is primitive. A binary polynomial of degree $m$ is *primitive* if it is irreducible and if the smallest positive integer $n$ such that the polynomial divides $X^n + 1$ is $n = 2^m - 1$.

        Returns:
            result: `True`, if the binary polynomial is primitive; `False`, otherwise.

        Examples:
            >>> komm.BinaryPolynomial(0b11011).is_primitive()
            False
            >>> komm.BinaryPolynomial(0b11111).is_primitive()
            False
            >>> komm.BinaryPolynomial(0b10011).is_primitive()
            True
        """
        if not self.is_irreducible():
            return False
        X = BinaryPolynomial(0b10)  # The polynomial X
        if self == X:
            return False
        order = 2**self.degree - 1
        for q in set(prime_factors(order)):
            if power_mod(X, order // q, self) == BinaryPolynomial(1):
                return False
        return True

    @classmethod
    def xgcd(cls, poly1: Self, poly2: Self) -> tuple[Self, Self, Self]:
        r"""
        Performs the extended Euclidean algorithm on two given binary polynomials.
        """
        return domain.xgcd(poly1, poly2)

    @classmethod
    def gcd(cls, *poly_list: Self) -> Self:
        r"""
        Computes the greatest common divisor (gcd) of the arguments.
        """
        return reduce(domain.gcd, poly_list)

    @classmethod
    def lcm(cls, *poly_list: Self) -> Self:
        r"""
        Computes the least common multiple (lcm) of the arguments.
        """
        return reduce(operator.mul, poly_list) // cls.gcd(*poly_list)


class BinaryPolynomials:
    def __call__(self, value: int) -> BinaryPolynomial:
        return BinaryPolynomial(value)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return True

    @property
    def zero(self) -> BinaryPolynomial:
        return BinaryPolynomial(0)

    @property
    def one(self) -> BinaryPolynomial:
        return BinaryPolynomial(1)


def default_primitive_polynomial(degree: int) -> BinaryPolynomial:
    # [LC04, Table 2.7, p. 47]
    if not 1 <= degree <= 24:
        raise ValueError("only degrees in the range [1 : 24] are implemented")
    return BinaryPolynomial(
        {
            1: 0b11,
            2: 0b111,
            3: 0b1011,
            4: 0b10011,
            5: 0b100101,
            6: 0b1000011,
            7: 0b10001001,
            8: 0b100011101,
            9: 0b1000010001,
            10: 0b10000001001,
            11: 0b100000000101,
            12: 0b1000001010011,
            13: 0b10000000011011,
            14: 0b100010001000011,
            15: 0b1000000000000011,
            16: 0b11010000000010001,
            17: 0b100000000000001001,
            18: 0b1000000000010000001,
            19: 0b10000000000000100111,
            20: 0b100000000000000001001,
            21: 0b1000000000000000000101,
            22: 0b10000000000000000000011,
            23: 0b100000000000000000100001,
            24: 0b1000000000000000010000111,
        }[degree]
    )


def power_mod(
    base: BinaryPolynomial, exponent: int, modulus: BinaryPolynomial
) -> BinaryPolynomial:
    # Computes (base ** exponent) % modulus using binary exponentiation.
    result = BinaryPolynomial(1)
    current = base
    while exponent > 0:
        if exponent & 1:
            result = (result * current) % modulus
        current = (current * current) % modulus
        exponent >>= 1
    return result



from functools import cache

import numpy as np
import numpy.typing as npt

from .. import abc
from ..types import Array1D, Array2D
from .PSKConstellation import PSKConstellation


class APSKConstellation(abc.Constellation[np.complexfloating]):
    r"""
    Amplitude- and phase-shift keying (APSK) constellation. It is a complex one-dimensional [constellation](/ref/Constellation) obtained by the union of $K$ component [PSK constellations](/ref/PSKConstellation), called *rings*.

    Parameters:
        orders: A $K$-tuple with the orders $M_k$ of each ring, for $k \in [0 : K)$.

        amplitudes: A $K$-tuple with the amplitudes $A_k$ of each ring, for $k \in [0 : K)$.

        phase_offsets: A $K$-tuple with the phase offsets $\phi_k$ of each ring, for $k \in [0 : K)$. If specified as a single float $\phi$, then it is assumed that $\phi_k = \phi$ for all $k \in [0 : K)$. The default value is `0.0`.

    Examples:
        1. The $8$-APSK constellation with $(M_0, M_1) = (4, 4)$, $(A_0, A_1) = (1, 2)$, and $(\phi_0, \phi_1) = (0, 0)$ is depicted below.
            <figure markdown>
            ![(4,4)-APSK constellation.](/fig/constellation_apsk_4_4.svg)
            </figure>

                >>> const = komm.APSKConstellation(orders=(4, 4), amplitudes=(1.0, 2.0))

        1. The $16$-APSK constellation with $(M_0, M_1) = (8, 8)$, $(A_0, A_1) = (1, 2)$, and $(\phi_0, \phi_1) = (0, 1/16)$ is depicted below.
            <figure markdown>
            ![(8,8)-APSK constellation.](/fig/constellation_apsk_8_8.svg)
            </figure>

                >>> const = komm.APSKConstellation(
                ...     orders=(8, 8),
                ...     amplitudes=(1.0, 2.0),
                ...     phase_offsets=(0.0, 1 / 16)
                ... )

        1. The $16$-APSK constellation with $(M_0, M_1) = (4, 12)$, $(A_0, A_1) = (\sqrt{2}, 3)$, and $(\phi_0, \phi_1) = (1/8, 0)$ is depicted below.
            <figure markdown>
            ![(4,12)-APSK constellation.](/fig/constellation_apsk_4_12.svg)
            </figure>

                >>> const = komm.APSKConstellation(
                ...     orders=(4, 12),
                ...     amplitudes=(np.sqrt(2), 3.0),
                ...     phase_offsets=(1 / 8, 0.0)
                ... )
    """

    _orders: tuple[int, ...]
    _amplitudes: tuple[float, ...]
    _phase_offsets: tuple[float, ...]

    def __init__(
        self,
        orders: tuple[int, ...],
        amplitudes: tuple[float, ...],
        phase_offsets: float | tuple[float, ...] = 0.0,
    ) -> None:
        self._orders = orders
        self._amplitudes = amplitudes
        if isinstance(phase_offsets, tuple):
            self._phase_offsets = phase_offsets
        else:
            self._phase_offsets = (phase_offsets,) * len(orders)

    def __repr__(self) -> str:
        args = ", ".join([
            f"orders={tuple(map(int, self._orders))}",
            f"amplitudes={tuple(map(float, self._amplitudes))}",
            f"phase_offsets={tuple(map(float, self._phase_offsets))}",
        ])
        return f"{self.__class__.__name__}({args})"

    @property
    @cache
    def matrix(self) -> Array2D[np.complexfloating]:
        r"""
        Examples:
            >>> const = komm.APSKConstellation(orders=(4, 4), amplitudes=(1.0, 2.0))
            >>> const.matrix
            array([[ 1.+0.j],
                   [ 0.+1.j],
                   [-1.+0.j],
                   [ 0.-1.j],
                   [ 2.+0.j],
                   [ 0.+2.j],
                   [-2.+0.j],
                   [ 0.-2.j]])
        """
        Ms, As, φs = self._orders, self._amplitudes, self._phase_offsets
        matrices = [PSKConstellation(M, A, φ).matrix for M, A, φ in zip(Ms, As, φs)]
        matrix = np.concatenate(matrices)
        return matrix.reshape(-1, 1)

    @property
    def order(self) -> int:
        r"""
        For the APSK constellation, it is given by
        $$
            M = \sum_{k \in [0:K)} M_k.
        $$

        Examples:
            >>> const = komm.APSKConstellation(orders=(4, 4), amplitudes=(1.0, 2.0))
            >>> const.order
            8
        """
        return sum(self._orders)

    @property
    def dimension(self) -> int:
        r"""
        For the APSK constellation, it is given by $N = 1$.

        Examples:
            >>> const = komm.APSKConstellation(orders=(4, 4), amplitudes=(1.0, 2.0))
            >>> const.dimension
            1
        """
        return 1

    def mean(self, priors: npt.ArrayLike | None = None) -> Array1D[np.complexfloating]:
        r"""
        For uniform priors, the mean of the APSK constellation is given by
        $$
            \mathbf{m} = 0.
        $$

        Examples:
            >>> const = komm.APSKConstellation(orders=(4, 4), amplitudes=(1.0, 2.0))
            >>> const.mean()
            array([0.+0.j])
        """
        if priors is None:
            return np.zeros((1,), dtype=complex)
        return super().mean(priors)

    def mean_energy(self, priors: npt.ArrayLike | None = None) -> np.floating:
        r"""
        For uniform priors, the mean energy of the APSK constellation is given by
        $$
            E = \frac{1}{M} \sum_{k \in [0:K)} A^2 M_k.
        $$

        Examples:
            >>> const = komm.APSKConstellation(orders=(4, 4), amplitudes=(1.0, 2.0))
            >>> const.mean_energy()
            np.float64(2.5)
        """
        if priors is None:
            Ms, As = self._orders, self._amplitudes
            return np.sum([M * A**2 for (M, A) in zip(Ms, As)]) / self.order
        return super().mean_energy(priors)

    def minimum_distance(self) -> np.floating:
        r"""
        Examples:
            >>> const = komm.APSKConstellation(orders=(4, 4), amplitudes=(1.0, 2.0))
            >>> const.minimum_distance()
            np.float64(1.0)
        """
        return super().minimum_distance()

    def indices_to_symbols(
        self, indices: npt.ArrayLike
    ) -> npt.NDArray[np.complexfloating]:
        r"""
        Examples:
            >>> const = komm.APSKConstellation(orders=(4, 4), amplitudes=(1.0, 2.0))
            >>> const.indices_to_symbols([3, 0])
            array([0.-1.j, 1.+0.j])
        """
        return super().indices_to_symbols(indices)

    def closest_indices(self, received: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> const = komm.APSKConstellation(orders=(4, 4), amplitudes=(1.0, 2.0))
            >>> const.closest_indices([0.1 - 1.1j, 1.2 + 0.1j])
            array([3, 0])
        """
        return super().closest_indices(received)

    def closest_symbols(
        self, received: npt.ArrayLike
    ) -> npt.NDArray[np.complexfloating]:
        r"""
        Examples:
            >>> const = komm.APSKConstellation(orders=(4, 4), amplitudes=(1.0, 2.0))
            >>> const.closest_symbols([0.1 - 1.1j, 1.2 + 0.1j])
            array([0.-1.j, 1.+0.j])
        """
        return super().closest_symbols(received)

    def posteriors(
        self,
        received: npt.ArrayLike,
        snr: float,
        priors: npt.ArrayLike | None = None,
    ) -> npt.NDArray[np.floating]:
        r"""
        Examples:
            >>> const = komm.APSKConstellation(orders=(4, 4), amplitudes=(1.0, 2.0))
            >>> const.posteriors([0.1 - 1.1j], snr=2.0).round(3)
            array([0.104, 0.015, 0.076, 0.516, 0.011, 0.   , 0.006, 0.272])
        """
        return super().posteriors(received, snr, priors)



import numpy as np
import numpy.typing as npt

from .. import abc


class Labeling(abc.Labeling):
    r"""
    General binary labeling. It is defined by a bijective mapping from $[0:2^m)$ to $\mathbb{B}^m$ or, equivalently, by an ordered set $\\{ \mathbf{q}_i : i \in [0:2^m) \\}$ of $2^m$ distinct binary vectors in  $\mathbb{B}^m$. In this class, a labeling is represented by a matrix $\mathbf{Q} \in \mathbb{B}^{2^m \times m}$, where the $i$-th row of $\mathbf{Q}$ corresponds to the binary vector $\mathbf{q}_i$. For more details, see <cite>SA15, Sec. 2.5.2</cite>.

    Parameters:
        matrix: The labeling matrix $\mathbf{Q}$. Must be a $2^m \times m$ binary matrix whose rows are all distinct.
    """

    def __init__(self, matrix: npt.ArrayLike) -> None:
        matrix = np.asarray(matrix)
        if matrix.ndim != 2:
            raise ValueError("'matrix' must be a 2D-array")
        M, m = matrix.shape
        if M != 2**m:
            raise ValueError(f"shape of 'matrix' must be (2**m, m) (got ({M}, {m}))")
        if np.any(matrix < 0) or np.any(matrix > 1):
            raise ValueError("elements of 'matrix' must be either 0 or 1")
        if len(set(tuple(row) for row in matrix)) != M:
            raise ValueError("rows of 'matrix' must be distinct")
        self._matrix = matrix

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.matrix.tolist()})"

    @property
    def matrix(self) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> labeling = komm.Labeling([[1, 0], [1, 1], [0, 1], [0, 0]])
            >>> labeling.matrix
            array([[1, 0],
                   [1, 1],
                   [0, 1],
                   [0, 0]])
        """
        return self._matrix

    @property
    def num_bits(self) -> int:
        r"""
        Examples:
            >>> labeling = komm.Labeling([[1, 0], [1, 1], [0, 1], [0, 0]])
            >>> labeling.num_bits
            2
        """
        return super().num_bits

    @property
    def cardinality(self) -> int:
        r"""
        Examples:
            >>> labeling = komm.Labeling([[1, 0], [1, 1], [0, 1], [0, 0]])
            >>> labeling.cardinality
            4
        """
        return super().cardinality

    @property
    def inverse_mapping(self) -> dict[tuple[int, ...], int]:
        r"""
        Examples:
            >>> labeling = komm.Labeling([[1, 0], [1, 1], [0, 1], [0, 0]])
            >>> labeling.inverse_mapping
            {(1, 0): 0, (1, 1): 1, (0, 1): 2, (0, 0): 3}
        """
        return super().inverse_mapping

    def indices_to_bits(self, indices: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> labeling = komm.Labeling([[1, 0], [1, 1], [0, 1], [0, 0]])
            >>> labeling.indices_to_bits([2, 0])
            array([0, 1, 1, 0])
            >>> labeling.indices_to_bits([[2, 0], [3, 3]])
            array([[0, 1, 1, 0],
                   [0, 0, 0, 0]])
        """
        return super().indices_to_bits(indices)

    def bits_to_indices(self, bits: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> labeling = komm.Labeling([[1, 0], [1, 1], [0, 1], [0, 0]])
            >>> labeling.bits_to_indices([0, 1, 1, 0])
            array([2, 0])
            >>> labeling.bits_to_indices([[0, 1, 1, 0], [0, 0, 0, 0]])
            array([[2, 0],
                   [3, 3]])
        """
        return super().bits_to_indices(bits)

    def marginalize(self, metrics: npt.ArrayLike) -> npt.NDArray[np.floating]:
        r"""
        Examples:
            >>> labeling = komm.Labeling([[1, 0], [1, 1], [0, 1], [0, 0]])
            >>> labeling.marginalize([0.1, 0.2, 0.3, 0.4, 0.25, 0.25, 0.25, 0.25])
            array([0.84729786, 0.        , 0.        , 0.        ])
            >>> labeling.marginalize([[0.1, 0.2, 0.3, 0.4], [0.25, 0.25, 0.25, 0.25]])
            array([[0.84729786, 0.        ],
                   [0.        , 0.        ]])
        """
        return super().marginalize(metrics)



from functools import cache

import numpy as np
import numpy.typing as npt

from .. import abc
from .._util.bit_operations import bits_to_int, int_to_bits


class ReflectedLabeling(abc.Labeling):
    r"""
    Reflected (Gray) binary labeling. It is a [binary labeling](/ref/Labeling) in which integer $i \in [0 : 2^m)$ is mapped to its Gray code representation.
    """

    def __init__(self, num_bits: int, _pre_cache: bool = True) -> None:
        if num_bits < 1:
            raise ValueError("'m' must be a positive integer")
        self._num_bits = num_bits
        if _pre_cache:
            self.matrix

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._num_bits})"

    @property
    @cache
    def matrix(self) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> labeling = komm.ReflectedLabeling(2)
            >>> labeling.matrix
            array([[0, 0],
                   [0, 1],
                   [1, 1],
                   [1, 0]])
        """
        m = self._num_bits
        ints = np.arange(2**m, dtype=int)
        return int_to_bits(ints ^ (ints >> 1), width=m, bit_order="MSB-first")

    @property
    def num_bits(self) -> int:
        r"""
        Examples:
            >>> labeling = komm.ReflectedLabeling(2)
            >>> labeling.num_bits
            2
        """
        return self._num_bits

    @property
    def cardinality(self) -> int:
        r"""
        Examples:
            >>> labeling = komm.ReflectedLabeling(2)
            >>> labeling.cardinality
            4
        """
        m = self._num_bits
        return 2**m

    @property
    @cache
    def inverse_mapping(self) -> dict[tuple[int, ...], int]:
        r"""
        Examples:
            >>> labeling = komm.ReflectedLabeling(2)
            >>> labeling.inverse_mapping
            {(0, 0): 0, (0, 1): 1, (1, 1): 2, (1, 0): 3}
        """
        return super().inverse_mapping

    def indices_to_bits(self, indices: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> labeling = komm.ReflectedLabeling(2)
            >>> labeling.indices_to_bits([2, 0])
            array([1, 1, 0, 0])
            >>> labeling.indices_to_bits([[2, 0], [3, 3]])
            array([[1, 1, 0, 0],
                   [1, 0, 1, 0]])
        """
        m = self._num_bits
        indices = np.asarray(indices, dtype=int)
        bits = int_to_bits(indices ^ indices >> 1, width=m, bit_order="MSB-first")
        return bits.reshape(*indices.shape[:-1], -1)

    def bits_to_indices(self, bits: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> labeling = komm.ReflectedLabeling(2)
            >>> labeling.bits_to_indices([1, 1, 0, 0])
            array([2, 0])
            >>> labeling.bits_to_indices([[1, 1, 0, 0], [1, 0, 1, 0]])
            array([[2, 0],
                   [3, 3]])
        """
        m = self._num_bits
        bits = np.asarray(bits, dtype=int)
        nat_indices = bits_to_int(bits.reshape(-1, m), bit_order="MSB-first")
        indices = np.zeros_like(nat_indices)
        for shift in range(m):
            indices ^= nat_indices >> shift
        return indices.reshape(*bits.shape[:-1], -1)

    def marginalize(self, metrics: npt.ArrayLike) -> npt.NDArray[np.floating]:
        r"""
        Examples:
            >>> labeling = komm.ReflectedLabeling(2)
            >>> labeling.marginalize([0.1, 0.2, 0.3, 0.4, 0.25, 0.25, 0.25, 0.25])
            array([-0.84729786,  0.        ,  0.        ,  0.        ])
            >>> labeling.marginalize([[0.1, 0.2, 0.3, 0.4], [0.25, 0.25, 0.25, 0.25]])
            array([[-0.84729786,  0.        ],
                   [ 0.        ,  0.        ]])
        """
        return super().marginalize(metrics)



from dataclasses import dataclass
from math import ceil, log

import numpy as np
import numpy.typing as npt
from tqdm import tqdm

from komm._util.validators import validate_integer_range

from .util import Word, integer_to_symbols, symbols_to_integer

Token = tuple[int, int]


@dataclass
class LempelZiv78Code:
    r"""
    Lempel–Ziv 78 (LZ78 or LZ2) code. It is a lossless data compression algorithm which is asymptotically optimal for ergodic sources. Let $\mathcal{X}$ be the source alphabet, and $\mathcal{Y}$ be the target alphabet. The token format is $(p, x)$, where $p \in \mathbb{N}$ is the index of the corresponding dictionary entry, and $x \in \mathcal{X}$ is the source symbol following the match. The index $p$ is represented as a variable-size word in $\mathcal{Y}^k$, where $k = \log_{|\mathcal{Y}|}(i + 1)$, and $i$ is the size of the dictionary at the moment. For more details, see <cite>MacK03, Sec. 6.4</cite>, <cite>Say06, Sec. 5.4.2</cite> and <cite>CT06, Sec. 13.4.2</cite>.

    Note:
        Here, for simplicity, we assume that the source alphabet is $\mathcal{X} = [0 : |\mathcal{X}|)$ and the target alphabet is $\mathcal{Y} = [0 : |\mathcal{Y}|)$, where $|\mathcal{X}| \geq 2$ and $|\mathcal{Y}| \geq 2$ are called the *source cardinality* and *target cardinality*, respectively.


    Parameters:
        source_cardinality: The source cardinality $|\mathcal{X}|$. Must satisfy $|\mathcal{X}| \geq 2$.
        target_cardinality: The target cardinality $|\mathcal{Y}|$. Must satisfy $|\mathcal{Y}| \geq 2$. The default value is $2$ (binary).

    Examples:
        >>> lz78 = komm.LempelZiv78Code(2)  # Binary source, binary target
        >>> lz78 = komm.LempelZiv78Code(3, 4)  # Ternary source, quaternary target
    """

    source_cardinality: int
    target_cardinality: int = 2

    def __post_init__(self) -> None:
        if not self.source_cardinality >= 2:
            raise ValueError("'source_cardinality' must be at least 2")
        if not self.target_cardinality >= 2:
            raise ValueError("'target_cardinality' must be at least 2")

    def source_to_tokens(self, source: npt.ArrayLike) -> list[Token]:
        r"""
        Encodes a given sequence of source symbols to the corresponding list of tokens.

        Examples:
            >>> lz78 = komm.LempelZiv78Code(2)
            >>> lz78.source_to_tokens([1, 0, 1, 1, 0, 1, 0, 1, 0, 0, 0])
            [(0, 1), (0, 0), (1, 1), (2, 1), (4, 0), (2, 0)]
        """
        source = validate_integer_range(source, high=self.source_cardinality)
        dictionary: dict[Word, int] = {(): 0}
        tokens: list[Token] = []
        word: Word = ()
        for symbol in tqdm(source, "Compressing LZ78", delay=2.5):
            if word + (symbol,) in dictionary:
                word += (symbol,)
                continue
            p, x = dictionary[word], int(symbol)
            tokens.append((p, x))
            dictionary[word + (symbol,)] = len(dictionary)
            word = ()
        if word:
            tokens.append((dictionary[word], -1))
        return tokens

    def tokens_to_source(self, tokens: list[Token]) -> npt.NDArray[np.integer]:
        r"""
        Decodes a given list of tokens to the corresponding sequence of source symbols.

        Examples:
            >>> lz78 = komm.LempelZiv78Code(2)
            >>> lz78.tokens_to_source([(0, 1), (0, 0), (1, 1), (2, 1), (4, 0), (2, 0)])
            array([1, 0, 1, 1, 0, 1, 0, 1, 0, 0, 0])
        """
        dictionary: dict[int, Word] = {0: ()}
        source: list[int] = []
        for p, x in tokens:
            word = dictionary[p]
            source.extend(word)
            if x >= 0:
                source.append(x)
            dictionary[len(dictionary)] = word + (x,)
        return np.array(source, dtype=int)

    def tokens_to_target(self, tokens: list[Token]) -> npt.NDArray[np.integer]:
        r"""
        Returns the target alphabet representation corresponding to a given list of tokens.

        Examples:
            >>> lz78 = komm.LempelZiv78Code(2)
            >>> lz78.tokens_to_target([(0, 1), (0, 0), (1, 1), (2, 1), (4, 0), (2, 0)])
            array([1, 0, 0, 0, 1, 1, 1, 0, 1, 1, 0, 0, 0, 0, 1, 0, 0])
        """
        calX, calY = self.source_cardinality, self.target_cardinality
        M = ceil(log(calX, calY))
        target: list[int] = []
        for i, (p, x) in enumerate(tokens):
            k = ceil(log(i + 1, calY))
            target.extend(integer_to_symbols(p, base=calY, width=k))
            if x >= 0:
                target.extend(integer_to_symbols(x, base=calY, width=M))
        return np.array(target, dtype=int)

    def target_to_tokens(self, target: npt.ArrayLike) -> list[Token]:
        r"""
        Returns the list of tokens corresponding to a given target alphabet representation.

        Examples:
            >>> lz78 = komm.LempelZiv78Code(2)
            >>> lz78.target_to_tokens([1, 0, 0, 0, 1, 1, 1, 0, 1, 1, 0, 0, 0, 0, 1, 0, 0])
            [(0, 1), (0, 0), (1, 1), (2, 1), (4, 0), (2, 0)]
        """
        calX, calY = self.source_cardinality, self.target_cardinality
        M = ceil(log(calX, calY))
        target = np.asarray(target)
        tokens: list[Token] = []
        i = 0
        while i < target.size:
            k = ceil(log(len(tokens) + 1, calY))
            p = int(symbols_to_integer(target[i : i + k], base=calY))
            i += k
            if i < target.size:
                x = int(symbols_to_integer(target[i : i + M], base=calY))
            else:
                x = -1
            tokens.append((p, x))
            i += M
        return tokens

    def encode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Encodes a sequence of source symbols to a sequence of target symbols.

        Parameters:
            input: The sequence of source symbols to be encoded. Must be a 1D-array with elements in $\mathcal{X}$.

        Returns:
            output: The sequence of encoded target symbols. It is a 1D-array with elements in $\mathcal{Y}$.

        Examples:
            >>> lz78 = komm.LempelZiv78Code(2)
            >>> lz78.encode([1, 0, 1, 1, 0, 1, 0, 1, 0, 0, 0])
            array([1, 0, 0, 0, 1, 1, 1, 0, 1, 1, 0, 0, 0, 0, 1, 0, 0])

            >>> lz78 = komm.LempelZiv78Code(2, 8)
            >>> lz78.encode(np.zeros(15, dtype=int))
            array([0, 1, 0, 2, 0, 3, 0, 4, 0])
        """
        tokens = self.source_to_tokens(input)
        output = self.tokens_to_target(tokens)
        return output

    def decode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Decodes a sequence of target symbols to a sequence of source symbols.

        Parameters:
            input: The sequence of target symbols to be decoded. Must be a 1D-array with elements in $\mathcal{Y}$. Also, the sequence must be a valid output of the `encode` method.

        Returns:
            output: The sequence of decoded source symbols. It is a 1D-array with elements in $\mathcal{X}$.

        Examples:
            >>> lz78 = komm.LempelZiv78Code(2)
            >>> lz78.decode([1, 0, 0, 0, 1, 1, 1, 0, 1, 1, 0, 0, 0, 0, 1, 0, 0])
            array([1, 0, 1, 1, 0, 1, 0, 1, 0, 0, 0])

            >>> lz78 = komm.LempelZiv78Code(2, 8)
            >>> lz78.decode([0, 1, 0, 2, 0, 3, 0, 4, 0])
            array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
        """
        tokens = self.target_to_tokens(input)
        output = self.tokens_to_source(tokens)
        return output



from dataclasses import dataclass
from functools import cache
from heapq import heapify, heappop, heappush
from itertools import product
from math import ceil, log2

import numpy as np
import numpy.typing as npt
from tqdm import tqdm
from typing_extensions import Self

from .._util.docs import mkdocstrings
from .._util.validators import validate_pmf
from ..types import Array1D
from .util import Word
from .VariableToFixedCode import VariableToFixedCode


@mkdocstrings(filters=["!.*"])
class TunstallCode(VariableToFixedCode):
    r"""
    Binary Tunstall code. It is an optimal (minimal expected rate) [variable-to-fixed length code](/ref/VariableToFixedCode) for a given pmf $p$ over $\mathcal{X}$. For more details, see <cite>Say06, Sec. 3.7</cite>.

    Notes:
        Tunstall codes are always [prefix-free](/ref/VariableToFixedCode/#is_prefix_free) (hence [uniquely encodable](/ref/VariableToFixedCode/#is_uniquely_encodable)) and [fully covering](/ref/VariableToFixedCode/#is_fully_covering).

    Parameters:
        pmf: The pmf $p$ to be considered. It must be a one-dimensional array of floats of size $|\mathcal{X}|$. The elements must be non-negative and sum to $1$.

        target_block_size: The target block size $n$. Must satisfy $2^n \geq |\mathcal{X}|$. The default value is $n = \lceil \log_2 |\mathcal{X}| \rceil$.

    Examples:
        >>> pmf = [0.8, 0.1, 0.1]

        >>> code = komm.TunstallCode(pmf)
        >>> code.dec_mapping
        {(0, 0): (0,),
         (0, 1): (1,),
         (1, 0): (2,)}
        >>> code.rate(pmf)  # doctest: +FLOAT_CMP
        np.float64(2.0)

        >>> code = komm.TunstallCode(pmf, 3)
        >>> code.dec_mapping
        {(0, 0, 0): (0, 0, 0),
         (0, 0, 1): (0, 0, 1),
         (0, 1, 0): (0, 0, 2),
         (0, 1, 1): (0, 1),
         (1, 0, 0): (0, 2),
         (1, 0, 1): (1,),
         (1, 1, 0): (2,)}
        >>> code.rate(pmf)  # doctest: +FLOAT_CMP
        np.float64(1.2295081967213108)
    """

    def __init__(
        self,
        pmf: npt.ArrayLike,
        target_block_size: int | None = None,
    ) -> None:
        self.pmf = validate_pmf(pmf)
        if target_block_size is None:
            target_block_size = ceil(log2(self.pmf.size))
        if 2**target_block_size < self.pmf.size:
            raise ValueError("'target_block_size' is too low")
        super().__init__(
            target_cardinality=2,
            source_cardinality=self.pmf.size,
            target_block_size=target_block_size,
            dec_mapping=tunstall_code(self.pmf, target_block_size),
        )

    def __repr__(self) -> str:
        args = ", ".join([
            f"pmf={self.pmf.tolist()}",
            f"target_block_size={self.target_block_size}",
        ])
        return f"{self.__class__.__name__}({args})"

    @cache
    def is_fully_covering(self) -> bool:
        return True

    @cache
    def is_uniquely_encodable(self) -> bool:
        return True

    @cache
    def is_prefix_free(self) -> bool:
        return True


def tunstall_code(
    pmf: Array1D[np.floating],
    target_block_size: int,
) -> dict[Word, Word]:

    @dataclass
    class Node:
        sourceword: Word
        probability: np.floating

        def __lt__(self, other: Self) -> np.bool:
            return -self.probability < -other.probability

    S, n = pmf.size, target_block_size
    K = ceil((2**n - S) / (S - 1)) - 1  # See [Say06, p. 70]
    size = S + K * (S - 1)

    pbar = tqdm(desc="Generating Tunstall code", total=K * S + size, delay=2.5)

    heap = [Node((symbol,), probability) for (symbol, probability) in enumerate(pmf)]
    heapify(heap)
    for _ in range(K):
        node = heappop(heap)
        for symbol, probability in enumerate(pmf):
            new_node = Node(node.sourceword + (symbol,), node.probability * probability)
            heappush(heap, new_node)
            pbar.update()

    dec_mapping: dict[Word, Word] = {}
    for y, x in zip(
        product([0, 1], repeat=n), sorted(node.sourceword for node in heap)
    ):
        dec_mapping[y] = x
        pbar.update()

    pbar.close()

    return dec_mapping
